from __future__ import annotations
import typing
from pathlib import Path
from urllib.parse import quote
import base64
from dataclasses import dataclass, field
from .env import env
from instaui.common.jsonable import dumps
from instaui.runtime.dataclass import JsLink

if typing.TYPE_CHECKING:
    from instaui.runtime.resource import StyleTag


_TCodeOrPath = typing.Union[str, Path]
_JS_PREFIX = "data:text/javascript;charset=utf-8"
_CSS_PREFIX = "data:text/css;charset=utf-8"
_ICON_PREFIX = "data:image/x-icon;base64"

_html_template = env.get_template("zero.html")


@dataclass(frozen=True)
class ZeroVueAppComponent:
    name: str
    url: _TCodeOrPath


@dataclass
class ZeroTemplateModel:
    version: str
    vue_js_code: _TCodeOrPath
    instaui_js_code: _TCodeOrPath
    config_dict: typing.Dict[str, typing.Any] = field(default_factory=dict)
    extra_import_maps: typing.Dict[str, _TCodeOrPath] = field(default_factory=dict)
    css_links: typing.List[_TCodeOrPath] = field(default_factory=list)
    style_tags: typing.List[StyleTag] = field(default_factory=list)
    js_links: typing.List[JsLink] = field(default_factory=list)
    script_tags: typing.List[str] = field(default_factory=list)
    vue_app_use: typing.List[str] = field(default_factory=list)
    vue_app_component: typing.List[ZeroVueAppComponent] = field(default_factory=list)
    prefix: str = ""
    title: typing.Optional[str] = None
    favicon: typing.Optional[Path] = None

    def add_extra_import_map(self, name: str, url: _TCodeOrPath):
        self.extra_import_maps[name] = url

    @property
    def import_maps_string(self):
        data = {
            **self.extra_import_maps,
            "vue": self.vue_js_code,
            "instaui": self.instaui_js_code,
        }

        return dumps(data)

    def normalize_path_with_self(self):
        self.vue_js_code = _normalize_path_to_dataurl(self.vue_js_code, _JS_PREFIX)
        self.instaui_js_code = _normalize_path_to_dataurl(
            self.instaui_js_code, _JS_PREFIX
        )

        self.css_links = [
            _normalize_path_to_dataurl(link, _CSS_PREFIX)
            for link in self.css_links
            if isinstance(link, str) or (isinstance(link, Path) and link.is_file())
        ]

        self.js_links = [
            JsLink(
                link=_normalize_path_to_dataurl(link.link, _JS_PREFIX),
                attrs=link.attrs,
            )
            for link in self.js_links
        ]

        self.extra_import_maps = {
            k: _normalize_path_to_dataurl(v, _JS_PREFIX)
            for k, v in self.extra_import_maps.items()
        }

        self.vue_app_component = [
            ZeroVueAppComponent(
                name=component.name,
                url=_normalize_path_to_dataurl(component.url, _JS_PREFIX),
            )
            for component in self.vue_app_component
        ]

        self.favicon = _normalize_path_to_base64_url(self.favicon, _ICON_PREFIX)  # type: ignore


def _normalize_path_to_dataurl(path: typing.Union[str, Path], prefix: str):
    if isinstance(path, Path):
        path = path.read_text(encoding="utf-8")

    return f"{prefix},{quote(path)}"


def _normalize_path_to_base64_url(path: typing.Optional[Path], prefix: str):
    if path is None:
        return None
    return f"{prefix},{base64.b64encode(path.read_bytes()).decode('utf-8')}"


def render_zero_html(model: ZeroTemplateModel) -> str:
    model.normalize_path_with_self()
    return _html_template.render(model=model)



from typing import Dict
from abc import ABC, abstractmethod


class PathableMixin(ABC):
    @abstractmethod
    def _to_pathable_binding_config(self) -> Dict:
        pass


class CanPathPropMixin(ABC):
    @abstractmethod
    def _to_path_prop_binding_config(self) -> Dict:
        pass



from __future__ import annotations
from enum import IntEnum
from typing import Literal, TypeVar


_T_Value = TypeVar("_T_Value")
_T_Output_Value = TypeVar("_T_Output_Value", covariant=True)
_T_VAR_TYPE = Literal["ref", "computed", "webComputed"]
_T_Bindable_Type = Literal["ref", "computed", "js", "webComputed", "vforItem"]


class InputBindingType(IntEnum):
    Ref = 0
    EventContext = 1
    Data = 2
    JsFn = 3


class OutputSetType(IntEnum):
    Ref = 0
    RouterAction = 1
    ElementRefAction = 2



from typing import Any, List, Sequence, TypedDict
from enum import IntEnum
from instaui.common.jsonable import Jsonable
from instaui.skip import is_skip_output
from instaui.patch_update import PatchSet


class TResponse(TypedDict, total=False):
    values: List[Any]
    types: Sequence[int]


class ValueType(IntEnum):
    VALUE = 0
    SKIP = 1
    Patch = 2


def response_data(outputs_binding_count: int, result: Any):
    data: TResponse = {}
    if outputs_binding_count > 0:
        if not isinstance(result, tuple):
            result = [result]

        returns_count = len(result)

        # [(value, 1), (value, 0)]
        result_infos = [
            (_try_get_value_from_jsonable(r), convert_type(r)) for r in result
        ]

        if returns_count == 1 and result_infos[0][1] == ValueType.SKIP:
            return data

        # fill missing values with None
        if returns_count < outputs_binding_count:
            result_infos.extend(
                [(None, ValueType.SKIP)] * (outputs_binding_count - returns_count)
            )

        data["values"] = [
            0 if info[1] == ValueType.SKIP else info[0] for info in result_infos
        ]
        types = [info[1] for info in result_infos]

        if sum(types) > 0:
            data["types"] = types

    return data


def convert_type(value: Any):
    if is_skip_output(value):
        return ValueType.SKIP
    if isinstance(value, PatchSet):
        return ValueType.Patch
    return ValueType.VALUE


def _try_get_value_from_jsonable(value: Any) -> Any:
    if isinstance(value, Jsonable):
        return value._to_json_dict()

    return value



import json


def py_value_to_js(value):
    if isinstance(value, str):
        return f"`{value}`"
    elif isinstance(value, (int, float)):
        return str(value)
    elif isinstance(value, bool):
        return "true" if value else "false"
    elif value is None:
        return "null"
    elif isinstance(value, dict):
        items = ", ".join(
            f"{json.dumps(k)}: {py_value_to_js(v)}" for k, v in value.items()
        )
        return f"{{{items}}}"
    elif isinstance(value, list):
        items = ", ".join(py_value_to_js(v) for v in value)
        return f"[{items}]"
    else:
        raise TypeError(f"Unsupported type: {type(value)}")



from __future__ import annotations

import ast
from copy import copy
import inspect
from pathlib import Path
import re
from typing import (
    Any,
    Callable,
    Dict,
    Iterable,
    List,
    ClassVar,
    Literal,
    Optional,
    Set,
    Tuple,
    Union,
    cast,
    overload,
    TYPE_CHECKING,
)
from typing_extensions import Self
from collections import defaultdict
from instaui.runtime._app import get_app_slot
from instaui.vars.element_ref import ElementRef
from instaui.vars.vfor_item import VForItem
from instaui.components.directive import Directive
from instaui.dependencies.component_dependency import (
    ComponentDependencyInfo,
)
from .slot import SlotManager, Slot
from instaui import consts
from instaui.components.component import Component

from instaui.vars.mixin_types.element_binding import (
    ElementBindingMixin,
    ElementBindingProtocol,
)
from instaui.vars.mixin_types.observable import ObservableMixin

if TYPE_CHECKING:
    from instaui.event.event_mixin import EventMixin
    from instaui.vars.types import TMaybeRef


TVarGetterStrategy = Union[Literal["as_needed", "all"], List]

# Refer to the NiceGUI project.
# https://github.com/zauberzeug/nicegui/blob/main/nicegui/element.py
PROPS_PATTERN = re.compile(
    r"""
# Match a key-value pair optionally followed by whitespace or end of string
([:\w\-]+)          # Capture group 1: Key
(?:                 # Optional non-capturing group for value
    =               # Match the equal sign
    (?:             # Non-capturing group for value options
        (           # Capture group 2: Value enclosed in double quotes
            "       # Match  double quote
            [^"\\]* # Match any character except quotes or backslashes zero or more times
            (?:\\.[^"\\]*)*  # Match any escaped character followed by any character except quotes or backslashes zero or more times
            "       # Match the closing quote
        )
        |
        (           # Capture group 3: Value enclosed in single quotes
            '       # Match a single quote
            [^'\\]* # Match any character except quotes or backslashes zero or more times
            (?:\\.[^'\\]*)*  # Match any escaped character followed by any character except quotes or backslashes zero or more times
            '       # Match the closing quote
        )
        |           # Or
        ([\w\-.%:\/]+)  # Capture group 4: Value without quotes
    )
)?                  # End of optional non-capturing group for value
(?:$|\s)            # Match end of string or whitespace
""",
    re.VERBOSE,
)


class Element(Component):
    dependency: ClassVar[Optional[ComponentDependencyInfo]] = None
    _default_props: ClassVar[Dict[str, Any]] = {}
    _default_classes: ClassVar[List[str]] = []
    _default_style: ClassVar[Dict[str, str]] = {}

    def __init__(self, tag: Optional[Union[str, ElementBindingProtocol]] = None):
        if self.dependency:
            tag = self.dependency.tag_name or ""

        super().__init__(tag)

        self._str_classes: List[str] = []
        self._dict_classes: Dict[str, ElementBindingMixin[bool]] = {}
        self._bind_str_classes: List[ElementBindingMixin[str]] = []
        self._str_classes.extend(self._default_classes)
        self._style: Dict[str, str] = {}
        self._style.update(self._default_style)
        self._style_str_binds: List[ElementBindingMixin[str]] = []
        self._props: Dict[str, Any] = {}
        self._props.update(self._default_props)
        self._proxy_props: List[ElementBindingMixin] = []

        self._events: defaultdict[str, List[EventMixin]] = defaultdict(list)
        self._directives: Dict[Directive, None] = {}

        self._slot_manager = SlotManager()
        self.__element_ref: Optional[ElementRef] = None
        self.__var_getter_strategy: TVarGetterStrategy = "as_needed"

    def __init_subclass__(
        cls,
        *,
        esm: Union[str, Path, None] = None,
        externals: Optional[Dict[str, Path]] = None,
        css: Union[List[Union[str, Path]], None] = None,
    ) -> None:
        super().__init_subclass__()

        if esm:
            esm = _make_dependency_path(esm, cls)

            if externals:
                externals = {
                    key: _make_dependency_path(value, cls)
                    for key, value in externals.items()
                }

            if css:
                css = set(_make_dependency_path(c, cls) for c in css)  # type: ignore

            tag_name = f"instaui-{esm.stem}"

            cls.dependency = ComponentDependencyInfo(
                tag_name=tag_name,
                esm=esm,
                externals=cast(Dict[str, Path], externals or {}),
                css=cast(Set[Path], css or set()),
            )

        cls._default_props = copy(cls._default_props)
        cls._default_classes = copy(cls._default_classes)
        cls._default_style = copy(cls._default_style)

    def on_mounted(
        self,
        handler: EventMixin,
        *,
        extends: Optional[List] = None,
    ):
        return self.on(
            ":mounted",
            handler=handler,
            extends=extends,
        )

    def _set_var_getter_strategy(self, strategy: TVarGetterStrategy):
        """Set the strategy for getting variables.

        # Example:
        .. code-block:: python
            self._set_var_getter_strategy("all")

            # or
            a = ui.state(1)
            b = ui.state(2)
            self._set_var_getter_strategy([a,b])

        """
        self.__var_getter_strategy = strategy

    def scoped_style(self, style: str, *, selector="*", with_self=False):
        app = get_app_slot()
        ssid = app.gen_scoped_style_group_id()

        select_box = f"*[insta-scoped-style={ssid}]"
        real_selector = f"{select_box} {selector}"

        if with_self:
            real_selector = f"{select_box},{real_selector}"

        real_selector = f":where({real_selector})"
        style_code = f"{real_selector} {{ {style} }}"

        self.props({"insta-scoped-style": ssid})
        app._html_resource.add_style_tag(
            style_code, group_id=consts.SCOPED_STYLE_GROUP_ID
        )
        return self

    def slot_props(self, name: str):
        return self._slot_manager.default.slot_props(name)

    @staticmethod
    def _update_classes(
        classes: List[str],
        add: str,
    ) -> List[str]:
        return list(dict.fromkeys(classes + add.split()))

    @staticmethod
    def _parse_style(text: Union[str, Dict[str, str]]) -> Dict[str, str]:
        if isinstance(text, dict):
            return text

        if not text:
            return {}

        result = {}
        for item in text.split(";"):
            item = item.strip()
            if item:
                key, value = item.split(":")
                key = key.strip()
                value = value.strip()
                result[key] = value

        return result

    @staticmethod
    def _parse_props(props: Union[str, Dict[str, Any]]) -> Dict[str, Any]:
        if isinstance(props, dict):
            return props

        if not props:
            return {}

        dictionary = {}
        for match in PROPS_PATTERN.finditer(props or ""):
            key = match.group(1)
            value = match.group(2) or match.group(3) or match.group(4)
            if value is None:
                dictionary[key] = True
            else:
                if (value.startswith("'") and value.endswith("'")) or (
                    value.startswith('"') and value.endswith('"')
                ):
                    value = ast.literal_eval(value)
                dictionary[key] = value
        return dictionary

    def key(self, key: Any):
        """Set the key prop of the component.

        Args:
            key (str): The key prop value.

        """
        self.props({"key": key})
        return self

    def vmodel(
        self,
        value: Any,
        modifiers: Union[consts.TModifier, List[consts.TModifier], None] = None,
        *,
        prop_name: str = "value",
        is_html_component=False,
    ):
        if prop_name == "value":
            prop_name = "modelValue"

        modifiers = modifiers or []
        if isinstance(modifiers, str):
            modifiers = [modifiers]

        self.directive(
            Directive(
                is_sys=is_html_component,
                name="vmodel",
                arg=prop_name,
                modifiers=modifiers,
                value=value,  # type: ignore
            )
        )

        if is_html_component is False:
            model_modifiers = {m: True for m in modifiers}
            self.props({"modelModifiers": model_modifiers})

        return self

    def add_slot(self, name: str) -> Slot:
        return self._slot_manager.get_slot(name)

    @overload
    def classes(self, add: str) -> Self: ...
    @overload
    def classes(self, add: Dict[str, TMaybeRef[bool]]) -> Self: ...

    @overload
    def classes(self, add: TMaybeRef[str]) -> Self: ...

    def classes(
        self,
        add: Union[
            str,
            Dict[str, TMaybeRef[bool]],
            TMaybeRef[str],
            VForItem,
        ],
    ) -> Self:
        """Add classes to the component.

        Args:
            add (Union[ str, Dict[str, TMaybeRef[bool]], TMaybeRef[str], VForItem, ]): classes to add.


        Examples:
        .. code-block:: python

            elemelt = html.span('test')
            elemelt.classes('class1 class2')

            # dynamically classes
            class_name = ui.state('x')
            elemelt.classes(class_name)

            # apply name if True
            apply = ui.state(True)
            elemelt.classes({'x': apply})
        """

        if isinstance(add, str):
            self._str_classes = self._update_classes(self._str_classes, add)

        if isinstance(add, dict):
            self._dict_classes.update(**add)  # type: ignore

        if isinstance(add, ElementBindingMixin):
            self._bind_str_classes.append(add)  # type: ignore

        return self

    def style(self, add: Union[str, Dict[str, Any], TMaybeRef[str]]) -> Self:
        if isinstance(add, dict):
            add = {key: value for key, value in add.items()}

        if isinstance(add, ElementBindingMixin):
            self._style_str_binds.append(add)
            return self

        new_style = self._parse_style(add)
        self._style.update(new_style)
        return self

    def props(self, add: Union[str, Dict[str, Any], TMaybeRef]) -> Self:
        if isinstance(add, ElementBindingMixin):
            self._proxy_props.append(add)
            return self

        if isinstance(add, dict):
            add = {key: value for key, value in add.items() if value is not None}

        new_props = self._parse_props(add)
        self._props.update(new_props)
        return self

    @classmethod
    def default_classes(cls, add: str) -> type[Self]:
        cls._default_classes = cls._update_classes(cls._default_classes, add)
        return cls

    @classmethod
    def default_style(cls, add: Union[str, Dict[str, str]]) -> type[Self]:
        new_style = cls._parse_style(add)
        cls._default_style.update(new_style)
        return cls

    @classmethod
    def default_props(cls, add: Union[str, Dict[str, Any]]) -> type[Self]:
        new_props = cls._parse_props(add)
        cls._default_props.update(new_props)
        return cls

    def on(
        self,
        event_name: str,
        handler: EventMixin,
        *,
        extends: Optional[List] = None,
    ):
        if extends:
            handler = handler.copy_with_extends(extends)

        self._events[event_name].append(handler)

        return self

    def directive(self, directive: Directive) -> Self:
        self._directives[directive] = None
        return self

    def display(self, value: Union[ElementBindingProtocol, bool]) -> Self:
        return self.directive(Directive(is_sys=False, name="vshow", value=value))

    def event_dataset(self, data: Any, name: str = "event-data") -> Self:
        from instaui.vars.js_computed import JsComputed

        value = JsComputed(inputs=[data], code="(data)=> JSON.stringify(data)")
        self.props({f"data-{name}": value})
        return self

    def element_ref(self, ref: ElementRef):
        self.__element_ref = ref
        return self

    def update_dependencies(
        self,
        *,
        css: Optional[Iterable[Path]] = None,
        externals: Optional[Dict[str, Path]] = None,
        replace: bool = False,
    ):
        if not self.dependency:
            return

        app = get_app_slot()
        dep = self.dependency.copy()
        if replace:
            dep.css.clear()
            dep.externals.clear()

        if css:
            dep.css.update(css)

        if externals:
            dep.externals.update(externals)

        app.add_temp_component_dependency(dep)

    def use(self, *use_fns: Callable[[Self], None]) -> Self:
        """Use functions to the component object.

        Args:
            use_fns (Callable[[Self], None]): The list of use functions.

        Examples:
        .. code-block:: python
            def use_red_color(element: html.paragraph):
                element.style('color: red')

            html.paragraph('Hello').use(use_red_color)
        """

        for fn in use_fns:
            fn(self)
        return self

    @classmethod
    def use_init(cls, init_fn: Callable[[type[Self]], Self]) -> Self:
        """Use this method to initialize the component.

        Args:
            init_fn (Callable[[type[Self]], Self]): The initialization function.

        Examples:
        .. code-block:: python
            def fack_init(cls: type[html.table]) -> html.table:
                return cls(columns=['name', 'age'],rows = [{'name': 'Alice', 'age': 25}, {'name': 'Bob', 'age': 30}])

            ui.table.use_init(fack_init)
        """
        return init_fn(cls)

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._style or self._style_str_binds:
            value_styles, bind_styles = _classifyBindableDict(self._style)
            if value_styles:
                data["style"] = value_styles

            b_style_list = []

            if bind_styles:
                b_style_list.append(bind_styles)

            if self._style_str_binds:
                b_style_list.append(
                    [v._to_element_binding_config() for v in self._style_str_binds]
                )

            if b_style_list:
                data["bStyle"] = b_style_list

        if self._str_classes or self._dict_classes or self._bind_str_classes:
            data["classes"] = _normalize_classes_data(
                self._str_classes, self._dict_classes, self._bind_str_classes
            )

        if self._props:
            value_props, bind_props = _classifyBindableDict(self._props)

            if value_props:
                data["props"] = value_props

            if bind_props:
                data["bProps"] = bind_props

        if self._proxy_props:
            data["proxyProps"] = [
                v._to_element_binding_config() for v in self._proxy_props
            ]

        if self._events:
            data["events"] = _normalize_events(self._events)

        if self._slot_manager.has_slot():
            data["slots"] = self._slot_manager

        if self._directives:
            data["dir"] = list(self._directives.keys())

        if self.dependency:
            app_slot = get_app_slot()
            tag_name = self.dependency.tag_name
            app_slot.use_component_dependency(
                app_slot.get_temp_component_dependency(tag_name, self.dependency)
            )

        if self.__element_ref:
            data["eRef"] = self.__element_ref._to_element_binding_config()

        if self.__var_getter_strategy != "as_needed":
            data["varGetterStrategy"] = (
                [
                    cast(ObservableMixin, v)._to_observable_config()
                    for v in self.__var_getter_strategy
                ]
                if isinstance(self.__var_getter_strategy, list)
                else self.__var_getter_strategy
            )

        return data


def _normalize_events(
    events: defaultdict[str, List[EventMixin]],
):
    return {
        _normalize_event_name(name): event_list for name, event_list in events.items()
    }


def _normalize_event_name(event_name: str):
    """'click' -> 'onClick' , 'press-enter' -> 'onPressEnter' , 'pressEnter' -> 'onPressEnter'"""

    if event_name.startswith("on-"):
        event_name = event_name[3:]

    if event_name.startswith("on"):
        event_name = event_name[2:]

    parts = event_name.split("-")
    formatted_parts = [part[0].upper() + part[1:] for part in parts]

    return "".join(["on", *formatted_parts])


def _normalize_classes_data(
    str_classes: List[str],
    dict_classes: Dict[str, ElementBindingMixin[bool]],
    bind_str_classes: List[ElementBindingMixin[str]],
):
    _str_result = " ".join(str_classes)

    _dict_classes = {k: v._to_element_binding_config() for k, v in dict_classes.items()}

    _bind_str_classes = [v._to_element_binding_config() for v in bind_str_classes]

    if _dict_classes or _bind_str_classes:
        result = {}

        if _str_result:
            result["str"] = _str_result

        if _dict_classes:
            result["map"] = _dict_classes

        if _bind_str_classes:
            result["bind"] = _bind_str_classes

        return result
    else:
        return _str_result


def _classifyBindableDict(
    data: Dict[str, Any],
) -> Tuple[Dict[str, Any], Dict[str, Any]]:
    """Return value_data, bind_data

    Args:
        data (Dict[str, Any]): _description_

    Returns:
        Tuple[Dict[str, Any], Dict[str, Any]]: _description_
    """

    value_data = {}
    bind_data = {}

    for key, value in data.items():
        if isinstance(value, ElementBindingMixin):
            bind_data[key] = value._to_element_binding_config()
        else:
            value_data[key] = value

    return value_data, bind_data


def _make_dependency_path(path: Union[str, Path], cls: type):
    if isinstance(path, str):
        path = Path(path)

    if not path.is_absolute():
        path = Path(inspect.getfile(cls)).parent / path

    return path



from __future__ import annotations
from typing import Callable, Dict, List, Optional
from instaui.consts import _T_App_Mode
from instaui.page_info import PageInfo
from instaui.systems import func_system


class LaunchCollector:
    _instance: Optional[LaunchCollector] = None

    @classmethod
    def get_instance(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = cls(*args, **kwargs)
        return cls._instance

    def __init__(self) -> None:
        self._page_router: Dict[str, PageInfo] = {}
        self._page_request_lifespans: Dict[int, Callable] = {}
        self._app_mode: _T_App_Mode = "web"

    @property
    def page_request_lifespans(self) -> List[Callable]:
        return list(self._page_request_lifespans.values())

    def register_page(self, info: PageInfo) -> None:
        self._page_router[info.path] = info

    def add_page_request_lifespan(self, lifespan: Callable) -> int:
        """Register a function to be called on each page request.

        Args:
            lifespan (Callable):  A function to be called on each page request.

        Returns:
            int:  A unique key to identify the registered lifespan function.
        """
        key = id(lifespan)
        self._page_request_lifespans[key] = func_system.make_fn_to_generator(lifespan)

        return key

    def remove_page_request_lifespan(self, key: int) -> None:
        if key in self._page_request_lifespans:
            del self._page_request_lifespans[key]

    def set_app_mode(self, mode: _T_App_Mode) -> None:
        self._app_mode = mode


def get_launch_collector() -> LaunchCollector:
    return LaunchCollector.get_instance()



from __future__ import annotations

from typing import TYPE_CHECKING, Dict, List, Optional
from instaui.common.jsonable import Jsonable
from instaui.runtime import get_slot_stacks, pop_slot
from instaui.runtime._app import ready_scope
from instaui.vars.slot_prop import BindingSlotPropItem

if TYPE_CHECKING:
    from instaui.components.component import Component

_DEFAULT_SLOT_NAME = ":"


class SlotManager(Jsonable):
    def __init__(self) -> None:
        super().__init__()
        self._slots: Dict[str, Slot] = {}

    def get_slot(self, name: str) -> Slot:
        name = _DEFAULT_SLOT_NAME if name == "default" else name

        if name not in self._slots:
            self._slots[name] = Slot(name)

        return self._slots[name]

    @property
    def default(self):
        return self.get_slot(_DEFAULT_SLOT_NAME)

    def _to_json_dict(self):
        return {name: slot._to_json_dict() for name, slot in self._slots.items()}

    def has_slot(self) -> bool:
        return len(self._slots) > 0 and any(
            slot._children for slot in self._slots.values()
        )


class Slot(Jsonable):
    def __init__(self, name: str) -> None:
        super().__init__()

        self._id: Optional[str] = None
        self._name = name
        self._children: List[Component] = []
        self._use_slot_props: bool = False
        self._scope_wrapper = ready_scope()
        self._slot_prop_info = SlotPropInfo(self._scope_wrapper)

    def slot_props(self, name: str):
        self._use_slot_props = True
        self._slot_prop_info.setup()

        item = BindingSlotPropItem(
            name, self._slot_prop_info.sid, self._slot_prop_info.var_id
        )

        return item[name]

    def __getitem__(self, item: str):
        return self.slot_props(item)

    def __enter__(self):
        get_slot_stacks().append(self)
        self._scope_wrapper.enter()

        return self

    def __exit__(self, *_):
        self._scope_wrapper.exit()
        pop_slot()

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._children:
            data["items"] = self._children

        if self._use_slot_props:
            data["use_prop"] = 1

        if self._scope_wrapper.used and self._scope_wrapper.scope.has_registered_task:
            data["sid"] = self._scope_wrapper.scope_id

        return data

    def _to_items_container_config(self) -> Dict:
        data = self._to_json_dict()

        return {k: v for k, v in data.items() if k in ("items", "sid")}


class SlotPropInfo(Jsonable):
    def __init__(self, scope_wrapper: ready_scope):
        pass

        self._scope_wrapper = scope_wrapper
        self.__register_info = None

    def setup(self):
        if self.__register_info is not None:
            return
        scope = self._scope_wrapper.scope
        self.__register_info = scope.register_slot_prop_info_task(self)

    @property
    def sid(self) -> str:
        return self.__register_info.scope_id  # type: ignore

    @property
    def var_id(self) -> str:
        return self.__register_info.var_id  # type: ignore

    def _to_json_dict(self):
        return {
            "sid": self.sid,
            "id": self.var_id,
        }



from typing import Protocol

import pydantic
import inspect


class TypeAdapterProtocol(Protocol):
    def to_python_value(self, value, *args, **kwargs): ...


class PydanticTypeAdapter:
    def __init__(self, type: type) -> None:
        self.type_adapter = pydantic.TypeAdapter(type)

    def to_python_value(self, value, *args, **kwargs):
        return self.type_adapter.validate_python(value)


class EmptyTypeAdapter:
    def to_python_value(self, value, *args, **kwargs):
        return value


def create_type_adapter(type: type) -> TypeAdapterProtocol:
    if type is inspect._empty:
        return EmptyTypeAdapter()
    return PydanticTypeAdapter(type)



from abc import ABC, abstractmethod
from typing import Sequence
from instaui.vars._types import InputBindingType, OutputSetType


class CanInputMixin(ABC):
    @abstractmethod
    def _to_input_config(self):
        pass

    @abstractmethod
    def _to_event_input_type(self) -> InputBindingType:
        pass


class CanOutputMixin(ABC):
    @abstractmethod
    def _to_output_config(self):
        pass

    @abstractmethod
    def _to_event_output_type(self) -> OutputSetType:
        pass


def inputs_to_config(inputs: Sequence[CanInputMixin]):
    return [
        {
            "value": input._to_input_config()
            if isinstance(input, CanInputMixin)
            else input,
            "type": input._to_event_input_type().value
            if isinstance(input, CanInputMixin)
            else InputBindingType.Data.value,
        }
        for input in inputs
    ]


def outputs_to_config(outputs: Sequence[CanOutputMixin]):
    return [
        {
            "ref": ref._to_output_config(),
            "type": ref._to_event_output_type().value,
        }
        for ref in outputs
    ]


def _assert_outputs_be_can_output_mixin(outputs: Sequence):
    for output in outputs:
        if not isinstance(output, CanOutputMixin):
            raise TypeError("The outputs parameter must be a `ui.state`")



from __future__ import annotations

from typing import (
    Dict,
    Optional,
    Union,
)
from instaui.common.jsonable import Jsonable
from instaui.runtime import (
    get_app_slot,
    check_default_app_slot_or_error,
)
from instaui.components.slot import SlotManager
from instaui.vars.mixin_types.element_binding import (
    ElementBindingMixin,
    ElementBindingProtocol,
)


class Component(Jsonable):
    def __init__(self, tag: Optional[Union[str, ElementBindingProtocol]] = None):
        check_default_app_slot_or_error(
            "Not allowed to create element outside of ui.page"
        )

        self._tag = (
            "div"
            if tag is None or tag == ""
            else (
                tag._to_element_binding_config()
                if isinstance(tag, ElementBindingMixin)
                else str(tag)
            )
        )
        self._slot_manager = SlotManager()

        get_app_slot().append_component_to_container(self)

    def __enter__(self):
        self._slot_manager.default.__enter__()
        return self

    def __exit__(self, *_) -> None:
        self._slot_manager.default.__exit__(*_)

    def _to_json_dict(self) -> Dict:
        data: Dict = {
            "tag": self._tag,
        }

        return data



__all__ = [
    "span",
    "label",
    "paragraph",
    "input",
    "number",
    "button",
    "checkbox",
    "radio",
    "form",
    "select",
    "option",
    "ul",
    "li",
    "div",
    "range",
    "date",
    "link",
    "textarea",
    "table",
    "h1",
    "h2",
    "h3",
    "h4",
    "h5",
    "h6",
    "image",
    "video",
]

from .span import Span as span
from .label import Label as label
from .paragraph import Paragraph as paragraph
from .input import Input as input
from .number import Number as number
from .button import Button as button
from .checkbox import Checkbox as checkbox
from .radio import Radio as radio
from .form import Form as form
from .select import Select as select
from .ul import Ul as ul
from .li import Li as li
from .div import Div as div
from .range import Range as range
from .date import Date as date
from .link import Link as link
from .textarea import Textarea as textarea
from .table import Table as table
from .heading import H1 as h1, H2 as h2, H3 as h3, H4 as h4, H5 as h5, H6 as h6
from .image import Image as image
from .video import Video as video

option = select.Option

from . import _preset  # noqa: E402, F401



from __future__ import annotations
import typing

from instaui.common.jsonable import Jsonable

from instaui.runtime._app import get_current_scope
from instaui.vars._types import InputBindingType
from instaui.vars.path_var import PathVar
from instaui.vars.mixin_types.var_type import VarMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.py_binding import CanInputMixin
from instaui.vars.mixin_types.pathable import CanPathPropMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.common_type import TObservableInput
from instaui._helper import observable_helper


class JsComputed(
    Jsonable,
    PathVar,
    VarMixin,
    ObservableMixin,
    CanInputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
):
    """
    A client-side computed property that evaluates JavaScript code to derive reactive values.

    Args:
        inputs (typing.Optional[typing.Sequence[TObservableInput]], optional): Reactive data sources to monitor.
                                                  Changes to these values trigger re-computation.
        code (str): JavaScript code to execute for computing the value.
        async_init_value (typing.Optional[typing.Any], optional): Initial value to use before first successful async evaluation.

    # Example:
    .. code-block:: python
        a = ui.state(0)

        plus_one = ui.js_computed(inputs=[a], code="a=> a+1")

        html.number(a)
        ui.label(plus_one)
    """

    def __init__(
        self,
        *,
        inputs: typing.Optional[typing.Sequence[TObservableInput]] = None,
        code: str,
        async_init_value: typing.Optional[typing.Any] = None,
        deep_compare_on_input: bool = False,
    ) -> None:
        self.code = code
        self._org_inputs = inputs or []

        scope = get_current_scope()
        self.__register_info = scope.register_js_computed_task(self)

        self._inputs, self._is_slient_inputs, self._is_data = (
            observable_helper.analyze_observable_inputs(list(inputs or []))
        )

        self._async_init_value = async_init_value
        self._deep_compare_on_input = deep_compare_on_input

    def __to_binding_config(self):
        return {
            "id": self.__register_info.var_id,
            "sid": self.__register_info.scope_id,
        }

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> typing.Dict:
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_pathable_binding_config(self) -> typing.Dict:
        return self.__to_binding_config()

    def _to_observable_config(self):
        return self.__to_binding_config()

    def _to_json_dict(self):
        data = super()._to_json_dict()

        data["id"] = self.__register_info.var_id

        if self._inputs:
            data["inputs"] = self._inputs

        if sum(self._is_slient_inputs) > 0:
            data["slient"] = self._is_slient_inputs

        if sum(self._is_data) > 0:
            data["data"] = self._is_data

        if self._async_init_value is not None:
            data["asyncInit"] = self._async_init_value

        if self._deep_compare_on_input is not False:
            data["deepEqOnInput"] = 1

        return data

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref


TJsComputed = JsComputed



from urllib.parse import urlparse
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.js_computed import JsComputed
from instaui.js.fn import JsFn

_complete_src_js_fn: JsFn = None  # type: ignore


def complete_src_computed(src: ElementBindingMixin):
    global _complete_src_js_fn

    if _complete_src_js_fn is None:
        _complete_src_js_fn = JsFn(
            code=r"""src=>{
            try {
                new URL(src);
                return src;
            } catch {
                if (!src.startsWith('/assets')) {
                src = src.startsWith('/')
                    ? '/assets' + src
                    : '/assets/' + src;
                }
                return src;
            }                   
            }""",
            global_scope=True,
        )

    return JsComputed(
        inputs=[src, _complete_src_js_fn],
        code=r"""(src,fn)=> fn(src)""",
    )


def complete_src(src: str) -> str:
    parsed = urlparse(src)

    if not parsed.scheme:
        if not src.startswith("/assets"):
            src = "/assets" + src if src.startswith("/") else "/assets/" + src
    return src



from abc import ABC, abstractmethod
import typing


class EventMixin(ABC):
    @abstractmethod
    def copy_with_extends(self, extends: typing.Sequence) -> "EventMixin":
        pass

    @abstractmethod
    def event_type(self) -> typing.Literal["web", "js"]:
        pass



from __future__ import annotations
from typing import (
    Dict,
    Generic,
    Optional,
    TypeVar,
    Union,
    overload,
)

from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_current_scope
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.path_var import PathVar

from .mixin_types.var_type import VarMixin
from .mixin_types.py_binding import CanInputMixin, CanOutputMixin
from .mixin_types.observable import ObservableMixin
from .mixin_types.element_binding import ElementBindingMixin
from .mixin_types.pathable import CanPathPropMixin
from .mixin_types.str_format_binding import StrFormatBindingMixin


_T_Value = TypeVar("_T_Value")


class Ref(
    Jsonable,
    PathVar,
    VarMixin,
    ObservableMixin,
    CanInputMixin,
    CanOutputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin[_T_Value],
    Generic[_T_Value],
):
    VAR_TYPE = "var"

    def __init__(
        self, value: Optional[_T_Value] = None, *, deep_compare: bool = False
    ) -> None:
        self.value = value  # type: ignore
        scope = get_current_scope()
        self.__register_info = scope.register_ref_task(self)

        self._deep_compare = deep_compare
        self._debounced = None

    def debounced(self, secounds: float):
        self._debounced = secounds
        return self

    def __to_binding_config(self):
        return {
            "id": self.__register_info.var_id,
            "sid": self.__register_info.scope_id,
        }

    def _to_pathable_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_observable_config(self):
        return self.__to_binding_config()

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_output_config(self):
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["id"] = self.__register_info.var_id

        if self._debounced is not None:
            data["debounced"] = self._debounced

        if self._deep_compare is True:
            data["deepCompare"] = True

        return data

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.Ref


@overload
def ref(value: Ref[_T_Value], *, deep_compare: bool = False) -> Ref[_T_Value]: ...


@overload
def ref(
    value: Optional[_T_Value] = None, *, deep_compare: bool = False
) -> Ref[_T_Value]: ...


def ref(
    value: Union[Ref[_T_Value], _T_Value, None] = None, *, deep_compare: bool = False
):
    if isinstance(value, Ref):
        return value
    return Ref(value, deep_compare=deep_compare)



from __future__ import annotations
from instaui.components.element import Element
from instaui.vars.types import TMaybeRef
from instaui.vars.js_computed import JsComputed


class Row(Element):
    def __init__(self, *, inline: TMaybeRef[bool] = False):
        super().__init__("div")
        flex = JsComputed(
            inputs=[inline], code="inline => inline? 'inline-flex' : 'flex'"
        )

        self.style("gap:var(--insta-row-gap)").style({"display": flex})

    def gap(self, value: str) -> Row:
        return self.style({"gap": value})



from ._app import (
    get_app_slot,
    reset_app_slot,
    new_app_slot,
    use_default_app_slot,
    get_slot_stacks,
    pop_slot,
    get_current_scope,
    ready_scope,
    check_default_app_slot_or_error,
    in_default_app_slot,
)
from ._inner_helper import check_web_only_mode_or_error
from .resource import HtmlResource

__all__ = [
    "get_slot_stacks",
    "ready_scope",
    "get_current_scope",
    "get_app_slot",
    "reset_app_slot",
    "new_app_slot",
    "use_default_app_slot",
    "pop_slot",
    "check_web_only_mode_or_error",
    "check_default_app_slot_or_error",
    "in_default_app_slot",
    "HtmlResource",
]



from pathlib import Path
from typing import Dict, Optional, Union
from instaui.runtime import new_app_slot, reset_app_slot
from contextlib import contextmanager
from .func import to_html, to_html_str, get_template_model
from instaui.template.zero_template import ZeroTemplateModel


@contextmanager
def scope(app_meta: Optional[Dict] = None):
    token = new_app_slot("zero", app_meta)
    yield Wrapper()
    reset_app_slot(token)


class Wrapper:
    def to_html(self, file: Union[str, Path]):
        file = self._get_caller_path(file)
        return to_html(file)

    def to_html_str(self):
        return to_html_str()

    def to_debug_report(self, file: Union[str, Path]):
        file = self._get_caller_path(file)

        # Custom component dependencies must be recorded only during actual execution
        result_html_str = to_html_str()
        model = get_template_model()
        with scope(app_meta={"merge_global_resources": False}) as s:
            _create_debug_report(model, result_html_str)
            s.to_html(file.resolve().absolute())

    @staticmethod
    def _get_caller_path(file: Union[str, Path]) -> Path:
        if isinstance(file, str):
            import inspect

            frame = inspect.currentframe().f_back.f_back  # type: ignore
            assert frame is not None
            script_file = inspect.getfile(frame)
            file = Path(script_file).parent.joinpath(file)

        return file


def _create_debug_report(model: ZeroTemplateModel, result_html_str: str):
    from instaui import ui, html

    no_exists_path_class = "ex-no-exists-path"

    def _path_exists_class(path: Path):
        return "" if path.exists() else no_exists_path_class

    ui.use_tailwind()

    ui.add_style(rf".{no_exists_path_class} {{background-color: red;color: white;}}")

    html_size = len(result_html_str.encode("utf-8")) / 1024 / 1024

    box_style = "border-2 border-gray-200 p-4 place-center gap-x-2"

    with ui.column().classes("gap-2"):
        # base info
        with ui.grid(columns="auto 1fr").classes(box_style):
            html.span("file size:")
            html.span(f"{html_size:.2f} MB")

        # import maps
        html.paragraph("import maps")

        rows = [
            ["vue", str(model.vue_js_code)],
            ["instaui", str(model.instaui_js_code)],
        ]

        for name, url in model.extra_import_maps.items():
            if isinstance(url, Path) and url.is_file():
                rows.append([name, str(url)])

        html.table(["name", "path"], rows).scoped_style(
            r"""
    table {
    outline: 1px solid black;
    }
    td, th {
    outline: 1px solid black;
    padding: 5px;
    }
    """,
            with_self=True,
        )

        # css links
        html.paragraph("css links")
        with ui.column().classes(box_style):
            for link in model.css_links:
                if isinstance(link, Path) and link.is_file():
                    html.span(str(link)).classes(_path_exists_class(link))

        # js links
        html.paragraph("js links")
        with ui.column().classes(box_style):
            for info in model.js_links:
                if isinstance(info.link, Path) and info.link.is_file():
                    html.span(str(info.link)).classes(_path_exists_class(info.link))

        # custom components
        html.paragraph("custom components")
        with ui.grid(columns="auto 1fr").classes(box_style):
            html.span("name")
            html.span("js file path")

            for info in model.vue_app_component:
                html.span(info.name)

                if isinstance(info.url, Path) and info.url.is_file():
                    html.span(str(info.url)).classes(_path_exists_class(info.url))
                else:
                    html.span("not file")



from __future__ import annotations
import typing
from pathlib import Path
from urllib.parse import quote
import base64


_JS_PREFIX = "data:text/javascript;charset=utf-8"
_CSS_PREFIX = "data:text/css;charset=utf-8"
_ICON_PREFIX = "data:image/x-icon;base64"


def normalize_path_to_dataurl(path: typing.Union[str, Path], prefix: str):
    if isinstance(path, Path):
        path = path.read_text(encoding="utf-8")

    return f"{prefix},{quote(path)}"


def normalize_path_to_base64_url(path: typing.Optional[Path], prefix: str):
    if path is None:
        return None
    return f"{prefix},{base64.b64encode(path.read_bytes()).decode('utf-8')}"



from __future__ import annotations
from typing import cast
from instaui.components.component import Component
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.types import TMaybeRef


class VIf(Component):
    def __init__(self, on: TMaybeRef[bool]):
        super().__init__("vif")
        self._on = cast(ElementBindingMixin, on)

    def _to_json_dict(self):
        data = super()._to_json_dict()

        data["on"] = (
            self._on
            if isinstance(self._on, bool)
            else self._on._to_element_binding_config()
        )

        if self._slot_manager.has_slot():
            slot_data = self._slot_manager.get_slot(
                "default"
            )._to_items_container_config()
            data.update(slot_data)

        return data



from typing import Dict, Tuple, TypeVar, cast
from instaui.vars.ref import Ref
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.pathable import CanPathPropMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.common.jsonable import Jsonable

from pydantic import BaseModel, RootModel

_T = TypeVar("_T")


_ProxyModel = RootModel


class RefProxy(
    CanInputMixin,
    ObservableMixin,
    CanOutputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
    Jsonable,
):
    def __init__(self, instance: BaseModel, *, deep_compare: bool = False) -> None:
        data = instance.model_dump()
        self._ref_ = Ref(data, deep_compare=deep_compare)
        self._prop_names_ = set(data.keys()) if isinstance(data, dict) else set()

    @property
    def __ref_(self):
        return super().__getattribute__("_ref_")

    def __getattribute__(self, name):
        if name not in super().__getattribute__("_prop_names_"):
            return super().__getattribute__(name)

        return self.__ref_[name]

    def __getitem__(self, name):
        return self.__ref_[name]

    def inverse(self):
        return self.__ref_.inverse()

    def __add__(self, other: str):
        return self.__ref_ + other

    def __radd__(self, other: str):
        return other + self.__ref_

    def _to_element_binding_config(self) -> Dict:
        return self.__ref_._to_element_binding_config()

    def _to_input_config(self):
        return self.__ref_._to_input_config()

    def _to_observable_config(self):
        return self.__ref_._to_observable_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__ref_._to_path_prop_binding_config()

    def _to_output_config(self):
        return self.__ref_._to_output_config()

    def _to_str_format_binding(self, order: int) -> Tuple[str, str]:
        return self.__ref_._to_str_format_binding(order)

    def _to_json_dict(self):
        return self.__ref_._to_json_dict()

    def _to_event_output_type(self) -> OutputSetType:
        return self.__ref_._to_event_output_type()

    def _to_event_input_type(self) -> InputBindingType:
        return self.__ref_._to_event_input_type()


class StateModel(BaseModel, Jsonable):
    pass

    def _to_json_dict(self):
        return self.model_dump()


def state(value: _T, *, deep_compare: bool = False) -> _T:
    """
    Creates a reactive state object that tracks changes and notifies dependencies.

    Args:
        value (_T): The initial value to wrap in a reactive state container.
                   Can be any type (primitive, object, or complex data structure).

    # Example:
    .. code-block:: python
        from instaui import ui,html
        count = ui.state(0)

        html.number(count)
        ui.label(count)
    """
    if isinstance(value, RefProxy):
        return value
    obj = RefProxy(_ProxyModel(value), deep_compare=deep_compare)
    return cast(_T, obj)



from __future__ import annotations
from contextlib import contextmanager
import inspect
import os
import multiprocessing
from pathlib import Path
from typing import Any, Optional, Set, Union
import __main__

from fastapi import FastAPI
from fastapi import Request
from fastapi.middleware.gzip import GZipMiddleware
from fastapi.staticfiles import StaticFiles
from fastapi.responses import HTMLResponse
import uvicorn
from uvicorn.supervisors import ChangeReload
import itertools

from instaui.html_tools import to_config_data
from instaui.launch_collector import get_launch_collector
from instaui.page_info import PageInfo

from instaui import consts
from instaui.runtime._app import get_app_slot, get_default_app_slot
from instaui.runtime.context import get_context
from instaui.runtime.dataclass import JsLink, VueAppComponent
from instaui.template import web_template
from instaui.runtime.resource import StyleTag


from . import dependency_router
from . import event_router
from . import watch_router
from . import debug_mode_router
from .middlewares import RequestContextMiddleware, NoCacheDebugModeMiddleware
from ._uvicorn import UvicornServer
from . import resource
from instaui.version import __version__ as _INSTA_VERSION

APP_IMPORT_STRING = "instaui.fastapi_server.server:Server._instance.app"


INSTAUI_STATIC_URL = f"/_instaui_{_INSTA_VERSION}/static"
INSTAUI_USER_ASSETS_URL = "/assets"


def _cal_static_link(target_path: Path):
    return f"{INSTAUI_STATIC_URL}/{target_path.relative_to(consts._STATIC_DIR)}"


VUE_JS_HASH_LINK = _cal_static_link(consts.VUE_ES_JS_PATH)
INSTAUI_JS_HASH_LINK = _cal_static_link(consts.APP_ES_JS_PATH)
APP_CSS_LINK = _cal_static_link(consts.APP_CSS_PATH)
FAVICON_LINK = _cal_static_link(consts.FAVICON_PATH)


class Server:
    _instance: Optional[Server] = None

    def __init__(
        self,
        *,
        caller_file_path: Path,
        use_gzip: Union[int, bool] = True,
    ):
        self._caller_file_path = caller_file_path
        self.app = FastAPI()
        self.app.add_middleware(RequestContextMiddleware)
        if get_context().debug_mode:
            self.app.add_middleware(NoCacheDebugModeMiddleware)

        if use_gzip:
            self.app.add_middleware(
                GZipMiddleware,
                minimum_size=use_gzip if isinstance(use_gzip, int) else 500,
            )

        dependency_router.create_router(self.app)
        event_router.create_router(self.app)
        watch_router.create_router(self.app)
        debug_mode_router.create_router(self.app)

        self.try_add_user_assets(self.app)
        self.add_instaui_static(self.app)

        for page_info in get_launch_collector()._page_router.values():
            self.register_page(page_info)

        self._registered_static_routes: Set[str] = set()

        Server._instance = self

    def register_page(self, info: PageInfo):
        is_async = inspect.iscoroutinefunction(info.func)

        self._remove_route(info.path)

        if is_async:

            @self.app.get(info.path)
            async def _(request: Request):
                self._update_page_info(request, info)
                with _execute_request_lifespans():
                    await info.func()
                html = self._to_web_html(
                    page_info=info,
                    request=request,
                )

                return HTMLResponse(html)

        else:

            @self.app.get(info.path)
            def _(request: Request):
                self._update_page_info(request, info)
                with _execute_request_lifespans():
                    info.func()
                html = self._to_web_html(
                    page_info=info,
                    request=request,
                )

                return HTMLResponse(html)

    def _to_web_html(
        self,
        *,
        page_info: PageInfo,
        request: Request,
    ):
        config_data = to_config_data()

        system_slot = get_app_slot()
        default_app_slot = get_default_app_slot()
        html_resource = system_slot._html_resource
        default_html_resource = default_app_slot._html_resource

        favicon_url = FAVICON_LINK
        if html_resource.favicon:
            favicon_url = resource.record_resource(html_resource.favicon)
        else:
            if default_html_resource.favicon:
                favicon_url = resource.record_resource(default_html_resource.favicon)

        model = web_template.WebTemplateModel(
            version=_INSTA_VERSION,
            vue_js_link=VUE_JS_HASH_LINK,
            instaui_js_link=INSTAUI_JS_HASH_LINK,
            css_links=[
                APP_CSS_LINK,
            ],
            config_dict=config_data,
            favicon_url=favicon_url,
            title=html_resource.title
            or default_html_resource.title
            or consts.PAGE_TITLE,
        )

        # register custom components
        for component in system_slot._component_dependencies:
            if not component.esm:
                continue

            model.vue_app_component.append(
                VueAppComponent(
                    name=component.tag_name,
                    url=resource.record_resource(component.esm),
                )
            )

            if component.css:
                for css_link in component.css:
                    css_resource = resource.record_resource(css_link)
                    if css_link.is_file():
                        model.css_links.append(css_resource)

            if component.externals:
                for name, url in component.externals.items():
                    model.add_extra_import_map(name, resource.record_resource(url))

        # register custom plugins
        for plugin in set(
            itertools.chain(
                system_slot._plugin_dependencies, default_app_slot._plugin_dependencies
            )
        ):
            if not plugin.esm:
                continue

            model.vue_app_use.append(plugin.name)

            model.add_extra_import_map(
                plugin.name, resource.record_resource(plugin.esm)
            )

            if plugin.css:
                for css_link in plugin.css:
                    model.css_links.append(resource.record_resource(css_link))

        # css file link to web static link
        for link in html_resource.get_valid_css_links(
            default_html_resource._css_links_manager
        ):
            if isinstance(link, Path):
                model.css_links.append(resource.record_resource(link))
            else:
                model.css_links.append(link)

        # js file link to web static link
        for info in html_resource.get_valid_js_links(
            default_html_resource._js_links_manager
        ):
            link = (
                resource.record_resource(info.link)
                if isinstance(info.link, Path)
                else info.link
            )
            model.js_links.append(JsLink(link))

        for js_code in itertools.chain(
            html_resource._script_tags, default_html_resource._script_tags
        ):
            model.script_tags.append(js_code)

        for sylte_tag in StyleTag.merge_by_group_id(
            itertools.chain(
                html_resource._style_tags, default_html_resource._style_tags
            )
        ):
            model.style_tags.append(sylte_tag)

        model.prefix = request.headers.get(
            "X-Forwarded-Prefix", request.scope.get("root_path", "")
        )

        return web_template.render_web_html(model)

    def _update_page_info(self, request: Request, page_info: PageInfo):
        app = get_app_slot()

        app._page_path = page_info.path
        app._page_params = request.path_params
        app._query_params = dict(request.query_params)

    def _remove_route(self, path: str) -> None:
        self.app.routes[:] = [
            r for r in self.app.routes if getattr(r, "path", None) != path
        ]

    def try_close_server(self):
        UvicornServer.instance.should_exit = True

    def run(
        self,
        host="0.0.0.0",
        port=8080,
        reload: bool = True,
        reload_dirs: str = ".",
        reload_includes: str = "*.py",
        reload_excludes: str = ".*, .py[cod], .sw.*, ~*",
        log_level="info",
        workers: int | None = None,
        uds: str | None = None,
        **kwargs: Any,
    ):
        if multiprocessing.current_process().name != "MainProcess":
            return

        if reload and not hasattr(__main__, "__file__"):
            reload = False

        config = uvicorn.Config(
            APP_IMPORT_STRING if reload else self.app,
            host=host,
            port=port,
            reload=reload,
            log_level=log_level,
            workers=workers,
            uds=uds,
            reload_includes=_split_args(reload_includes) if reload else None,
            reload_excludes=_split_args(reload_excludes) if reload else None,
            reload_dirs=_split_args(reload_dirs) if reload else None,
            **kwargs,
        )

        UvicornServer.create_singleton(config, [debug_mode_router.when_server_reload])

        if config.should_reload:
            ChangeReload(config, target=UvicornServer.instance.run, sockets=[]).run()
        else:
            UvicornServer.instance.run()

        if config.uds:
            os.remove(config.uds)  # pragma: py-win32

    def run_with(self, app):
        assert isinstance(app, FastAPI), "app must be a FastAPI instance"

    @staticmethod
    def add_instaui_static(app: FastAPI):
        app.mount(
            INSTAUI_STATIC_URL,
            StaticFiles(directory=consts._STATIC_DIR),
            name=INSTAUI_STATIC_URL,
        )

    def try_add_user_assets(self, app: FastAPI):
        assets_dir = self._caller_file_path.parent.joinpath("assets")
        if not assets_dir.exists():
            return

        app.mount(INSTAUI_USER_ASSETS_URL, StaticFiles(directory=assets_dir))

        print(
            f"User assets directory found: {assets_dir} : mount to {INSTAUI_USER_ASSETS_URL}"
        )


def _split_args(args: str):
    return [a.strip() for a in args.split(",")]


@contextmanager
def _execute_request_lifespans():
    events = [iter(event()) for event in get_launch_collector().page_request_lifespans]
    for event in events:
        next(event)

    yield

    for event in events:
        try:
            next(event)
        except StopIteration:
            pass



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Union
from instaui.components.element import Element

if TYPE_CHECKING:
    from instaui.vars.types import TMaybeRef


class Paragraph(Element):
    """
    A component class representing an HTML `<p>` (paragraph) element.

    Args:
        text (Union[str, TMaybeRef[Any]]):The text content of the paragraph.
                                          - If a string is provided, the content is static.
                                          - If a `TMaybeRef` object is provided, the content
                                            will reactively update when the referenced value changes.
    """

    def __init__(
        self,
        text: Union[str, TMaybeRef[Any]],
    ):
        super().__init__("p")
        self.props(
            {
                "innerText": text,
            }
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Callable, Dict, List, Optional
import functools
import weakref
from instaui.common.jsonable import Jsonable


if TYPE_CHECKING:
    from instaui.vars.mixin_types.var_type import VarMixin
    from instaui.vars.mixin_types.py_binding import CanInputMixin
    from instaui.vars.web_computed import WebComputed
    from instaui.vars.js_computed import JsComputed
    from instaui.vars.vue_computed import VueComputed
    from instaui.vars.data import ConstData
    from instaui.watch.web_watch import WebWatch
    from instaui.watch.js_watch import JsWatch
    from instaui.watch.vue_watch import VueWatch
    from instaui.vars.element_ref import ElementRef
    from instaui.vars.vfor_item import VForItem, VForIndex, VForItemKey
    from instaui.components.slot import SlotPropInfo
    from instaui.js.fn import JsFn


class Scope(Jsonable):
    def __init__(self, id: str) -> None:
        super().__init__()
        self.id = id
        self._vars_id_counter = 0
        self._element_ref_id_counter = 0
        self._refs: List[VarMixin] = []
        self._element_refs: List[ElementRef] = []
        self._const_data: List[ConstData] = []
        self._js_computeds: List[JsComputed] = []
        self._vue_computeds: List[VueComputed] = []
        self._web_computeds: List[WebComputed] = []
        self._run_method_records: List = []
        self._web_watch_configs: List[Dict] = []
        self._js_watch_configs: List[Dict] = []
        self._vue_watch_configs: List[Dict] = []
        self._vfor_item: Optional[VForItem] = None
        self._vfor_item_key: Optional[VForItemKey] = None
        self._vfor_index: Optional[VForIndex] = None
        self._slot_prop_info: Optional[SlotPropInfo] = None
        self._js_fns: List[JsFn] = []
        self._query = {}
        self.__has_registered_task = False

    def _mark_has_registered_task(self):
        self.__has_registered_task = True

    @property
    def has_var(self):
        return self._vars_id_counter > 0

    @property
    def has_registered_task(self):
        return self.__has_registered_task

    def generate_vars_id(self) -> str:
        self._vars_id_counter += 1
        return str(self._vars_id_counter)

    def set_query(self, url: str, key: str, on: List[CanInputMixin]) -> None:
        self._query = {
            "url": url,
            "key": key,
            "on": [v._to_input_config() for v in on],
        }

    def register_web_watch(self, watch: WebWatch) -> None:
        self._mark_has_registered_task()
        self._web_watch_configs.append(watch._to_json_dict())

    def register_js_watch(self, watch: JsWatch) -> None:
        self._mark_has_registered_task()
        self._js_watch_configs.append(watch._to_json_dict())

    def register_vue_watch(self, watch: VueWatch) -> None:
        self._mark_has_registered_task()
        self._vue_watch_configs.append(watch._to_json_dict())

    def register_data_task(self, data: ConstData):
        weak_obj = weakref.ref(data)

        def register_fn():
            self._const_data.append(weak_obj())  # type: ignore
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def register_ref_task(self, ref: VarMixin):
        weak_obj = weakref.ref(ref)

        def register_fn():
            self._refs.append(weak_obj())  # type: ignore
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def register_element_ref_task(self, target: ElementRef):
        weak_obj = weakref.ref(target)

        def register_fn():
            self._element_refs.append(weak_obj())  # type: ignore
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def register_vfor_item_task(self, vfor_item: VForItem):
        weak_obj = weakref.ref(vfor_item)

        def register_fn():
            self._vfor_item = weak_obj()
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def register_slot_prop_info_task(self, slot_prop_info: SlotPropInfo):
        weak_obj = weakref.ref(slot_prop_info)

        def register_fn():
            self._slot_prop_info = weak_obj()
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def register_vfor_index_task(self, vfor_index: VForIndex):
        weak_obj = weakref.ref(vfor_index)

        def register_fn():
            self._vfor_index = weak_obj()
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def register_vfor_key_task(self, vfor_key: VForItemKey):
        weak_obj = weakref.ref(vfor_key)

        def register_fn():
            self._vfor_item_key = weak_obj()
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def register_js_computed_task(self, computed: JsComputed):
        weak_obj = weakref.ref(computed)

        def register_fn():
            self._js_computeds.append(weak_obj())  # type: ignore
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def register_computed_task(self, computed: WebComputed):
        weak_obj = weakref.ref(computed)

        def register_fn():
            self._web_computeds.append(weak_obj())  # type: ignore
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def register_vue_computed_task(self, computed: VueComputed):
        weak_obj = weakref.ref(computed)

        def register_fn():
            self._vue_computeds.append(weak_obj())  # type: ignore
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def register_js_fn_task(self, fn: JsFn):
        weak_obj = weakref.ref(fn)

        def register_fn():
            self._js_fns.append(weak_obj())  # type: ignore
            return self.generate_vars_id()

        return VarRegisterTaskForScope(self, register_fn)

    def _to_json_dict(self):
        data = super()._to_json_dict()
        if self._refs:
            data["refs"] = self._refs
        if self._query:
            data["query"] = self._query
        if self._web_watch_configs:
            data["py_watch"] = self._web_watch_configs
        if self._js_watch_configs:
            data["js_watch"] = self._js_watch_configs
        if self._vue_watch_configs:
            data["vue_watch"] = self._vue_watch_configs
        if self._element_refs:
            data["eRefs"] = self._element_refs

        if self._web_computeds:
            data["web_computed"] = self._web_computeds

        if self._js_computeds:
            data["js_computed"] = self._js_computeds

        if self._vue_computeds:
            data["vue_computed"] = self._vue_computeds
        if self._const_data:
            data["data"] = self._const_data

        if self._js_fns:
            data["jsFn"] = self._js_fns

        # vfor info
        vfor_dict = {}
        if self._vfor_item:
            vfor_dict[self._vfor_item.SCOPE_TYPE] = self._vfor_item
        if self._vfor_index:
            vfor_dict[self._vfor_index.SCOPE_TYPE] = self._vfor_index
        if self._vfor_item_key:
            vfor_dict[self._vfor_item_key.SCOPE_TYPE] = self._vfor_item_key

        if vfor_dict:
            data["vfor"] = vfor_dict

        if self._slot_prop_info:
            data["sp"] = self._slot_prop_info._to_json_dict()

        return data


class GlobalScope(Scope):
    def __init__(self, id: str) -> None:
        super().__init__(id)

    def register_ref_task(self, var: VarMixin) -> None:
        raise ValueError("Can not register ref in global scope")

    def register_computed_task(self, computed: WebComputed) -> None:
        raise ValueError("Can not register web_computeds  in global scope")

    def register_js_computed_task(self, computed: JsComputed):
        raise ValueError("Can not register js_computeds  in global scope")

    def register_vue_computed_task(self, computed: VueComputed):
        raise ValueError("Can not register vue_computeds  in global scope")

    def register_web_watch(self, watch: WebWatch) -> None:
        raise ValueError("Can not register web_watchs  in global scope")

    def register_js_watch(self, watch: JsWatch) -> None:
        raise ValueError("Can not register js_watchs  in global scope")

    def register_vue_watch(self, watch: VueWatch) -> None:
        raise ValueError("Can not register vue_watchs  in global scope")


class VarRegisterTaskForScope:
    def __init__(self, scope: Scope, register_fn: Callable[[], str]) -> None:
        scope._mark_has_registered_task()
        self._scope_id = scope.id
        self._id_gen_fn = functools.lru_cache(maxsize=1)(register_fn)

    @property
    def scope_id(self) -> str:
        return self._scope_id

    @property
    def var_id(self) -> str:
        return self._id_gen_fn()



from __future__ import annotations
from typing import (
    Any,
    Dict,
)

from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_current_scope
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.path_var import PathVar

from .mixin_types.var_type import VarMixin
from .mixin_types.py_binding import CanInputMixin, CanOutputMixin
from .mixin_types.element_binding import ElementBindingMixin
from .mixin_types.pathable import CanPathPropMixin
from .mixin_types.str_format_binding import StrFormatBindingMixin


class ConstData(
    Jsonable,
    PathVar,
    VarMixin,
    CanInputMixin,
    CanOutputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
):
    def __init__(self, value: Any = None) -> None:
        self.value = value  # type: ignore

        scope = get_current_scope()
        self.__register_info = scope.register_data_task(self)

    def __to_binding_config(self):
        return {
            "id": self.__register_info.var_id,
            "sid": self.__register_info.scope_id,
        }

    def _to_pathable_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_output_config(self):
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["id"] = self.__register_info.var_id

        return data

    def _to_event_output_type(self) -> OutputSetType:
        raise TypeError("ConstData cannot be used as an output")

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Data


TConstData = ConstData



from typing import Any, TypeVar, Union
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin

_T = TypeVar("_T")


def is_bindable(obj: Any):
    return isinstance(
        obj, (CanInputMixin, CanOutputMixin, ObservableMixin, ElementBindingMixin)
    )


TBindable = Union[
    CanInputMixin, CanOutputMixin, ObservableMixin, ElementBindingMixin[_T]
]



import contextvars
from typing import Dict


_scope_var: contextvars.ContextVar[Dict[type, object]] = contextvars.ContextVar(
    "_scope_var", default={}
)


def save_state(key, obj: object) -> None:
    _scope_var.get()[key] = obj


def load_state(key) -> object:
    return _scope_var.get()[key]



from fastapi import FastAPI
from fastapi.responses import FileResponse

from instaui.fastapi_server import resource


URL = f"{resource.URL}/{{hash_part:path}}/{{file_name:path}}"


def create_router(app: FastAPI):
    _dependency_handler(app)


def _dependency_handler(app: FastAPI):
    @app.get(URL)
    def _(hash_part: str, file_name: str) -> FileResponse:
        hash_part_with_extend_paths = hash_part.split("/", maxsplit=1)
        hash_part = hash_part_with_extend_paths[0]
        extend_path = (
            None
            if len(hash_part_with_extend_paths) == 1
            else hash_part_with_extend_paths[1]
        )

        folder = resource.get_folder_path(hash_part)
        if extend_path:
            folder = folder.joinpath(extend_path)
        local_file = folder / file_name

        return FileResponse(
            local_file, headers={"Cache-Control": "public, max-age=3600"}
        )



from __future__ import annotations

from typing import TYPE_CHECKING, Any, ClassVar, Dict, List, Optional, Set
from typing_extensions import Unpack, TypedDict
from instaui.common.jsonable import Jsonable
from .resource import HtmlResource
from instaui.consts import _T_App_Mode
from contextvars import ContextVar, Token
from instaui.runtime.scope import Scope, GlobalScope
from types import MappingProxyType

if TYPE_CHECKING:
    from instaui.components.component import Component
    from instaui.components.slot import Slot

    from instaui.dependencies.component_dependency import ComponentDependencyInfo
    from instaui.dependencies.plugin_dependency import PluginDependencyInfo
    from instaui.spa_router._route_model import RouteCollector


class App(Jsonable):
    _default_app_slot: ClassVar[Optional[App]] = None
    _web_server_info: ClassVar[Optional[Dict]] = None

    def __init__(self, *, mode: _T_App_Mode, meta: Optional[Dict] = None) -> None:
        super().__init__()
        self._scope_id_counter = 0
        self._scoped_style_id_counter = 0
        self._mode: _T_App_Mode = mode
        self.items: List[Component] = []
        self.meta = meta
        self._slots_stacks: List[Slot] = []

        defalut_scope = self.create_scope()
        self._scope_stack: List[Scope] = [defalut_scope]
        self._scopes: List[Scope] = [defalut_scope]
        self._sid = defalut_scope.id
        self._html_resource = HtmlResource()
        self._component_dependencies: Set[ComponentDependencyInfo] = set()
        self._temp_component_dependencies: Dict[str, ComponentDependencyInfo] = {}
        self._plugin_dependencies: Set[PluginDependencyInfo] = set()

        self._page_path: Optional[str] = None
        self._page_params: Dict[str, Any] = {}
        self._query_params: Dict[str, Any] = {}
        self._route_collector: Optional[RouteCollector] = None

    @property
    def mode(self) -> _T_App_Mode:
        return self._mode

    @property
    def top_scope(self) -> Scope:
        return self._scope_stack[0]

    @property
    def page_path(self) -> str:
        assert self._page_path is not None, "Page path is not set"
        return self._page_path  # type: ignore

    @property
    def page_params(self):
        return MappingProxyType(self._page_params)

    @property
    def query_params(self):
        return MappingProxyType(self._query_params)

    def create_scope(self) -> Scope:
        self._scope_id_counter += 1
        scope = Scope(str(self._scope_id_counter))
        return scope

    def gen_scoped_style_group_id(self):
        gid = f"scoped-style-{self._scoped_style_id_counter}"
        self._scoped_style_id_counter += 1
        return gid

    def reset_html_resource(self):
        self._html_resource = HtmlResource()

    def add_temp_component_dependency(self, dependency: ComponentDependencyInfo):
        self._temp_component_dependencies[dependency.tag_name] = dependency

    def get_temp_component_dependency(
        self, tag_name: str, default: ComponentDependencyInfo
    ) -> ComponentDependencyInfo:
        return self._temp_component_dependencies.get(tag_name, default)

    def has_temp_component_dependency(self, tag_name: str):
        return tag_name in self._temp_component_dependencies

    def use_component_dependency(
        self, dependency: ComponentDependencyInfo, *, replace=False
    ) -> None:
        if replace:
            self._component_dependencies.discard(dependency)

        self._component_dependencies.add(dependency)

    def use_plugin_dependency(self, dependency: PluginDependencyInfo) -> None:
        self._plugin_dependencies.add(dependency)

    def register_router(self, collector: RouteCollector) -> None:
        self._route_collector = collector

    def append_component_to_container(self, component: Component):
        if self._slots_stacks:
            self._slots_stacks[-1]._children.append(component)
        else:
            self.items.append(component)

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._page_path:
            url_info = {"path": self.page_path}
            if self._page_params:
                url_info["params"] = self._page_params  # type: ignore

            data["url"] = url_info

        # data["scopes"] = self._scopes

        if self._route_collector is not None:
            data["router"] = self._route_collector.model_dump(
                exclude_defaults=True, by_alias=True
            )

        if self._web_server_info is not None:
            data["webInfo"] = self._web_server_info

        data["sid"] = self._sid

        return data

    @classmethod
    def _create_default(cls):
        if cls._default_app_slot is None:
            cls._default_app_slot = DefaultApp(mode="web")
        return cls._default_app_slot


class DefaultApp(App):
    _instance = None

    def __new__(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = super(DefaultApp, cls).__new__(cls)
        return cls._instance

    def create_scope(self) -> Scope:
        self._scope_id_counter += 1
        scope = GlobalScope(str(self._scope_id_counter))
        return scope

    def append_component_to_container(self, component: Component):
        raise ValueError("Operations are not allowed outside of ui.page")


_app_var: ContextVar[App] = ContextVar("_app_var", default=App._create_default())


def use_default_app_slot():
    assert App._default_app_slot is not None, "Default app slot is not set"
    _app_var.set(App._default_app_slot)


def get_default_app_slot():
    return App._create_default()


def get_app_slot() -> App:
    return _app_var.get()


def get_current_scope():
    current_scope = get_app_slot()._scope_stack[-1]
    if current_scope is None:
        raise ValueError("No current scope")
    return current_scope


class ready_scope:
    def __init__(self) -> None:
        self._scope = None

    @property
    def used(self) -> bool:
        return self._scope is not None

    @property
    def has_vars(self) -> bool:
        return self._scope.has_var  # type: ignore

    @property
    def scope(self) -> Scope:
        return self._scope  # type: ignore

    @property
    def scope_id(self) -> str:
        return self._scope.id  # type: ignore

    def __enter__(self):
        return self.enter()

    def __exit__(self, *_) -> None:
        self.exit(*_)

    def enter(self):
        if self._scope is not None:
            raise ValueError("Scope is already used")

        app = get_app_slot()
        self._scope = app.create_scope()
        app._scopes.append(self._scope)
        app._scope_stack.append(self._scope)

        return self._scope

    def exit(self, *_) -> None:
        get_app_slot()._scope_stack.pop()


def get_slot_stacks():
    return get_app_slot()._slots_stacks


def pop_slot():
    get_slot_stacks().pop()


def new_app_slot(mode: _T_App_Mode, app_meta: Optional[Dict] = None):
    return _app_var.set(App(mode=mode, meta=app_meta))


def reset_app_slot(token: Token[App]):
    _app_var.reset(token)


def in_default_app_slot():
    return isinstance(get_app_slot(), DefaultApp)


def check_default_app_slot_or_error(
    error_message="Operations are not allowed outside of ui.page",
):
    if isinstance(get_app_slot(), DefaultApp):
        raise ValueError(error_message)


class TWebServerInfo(TypedDict, total=False):
    watch_url: Optional[str]
    watch_async_url: Optional[str]
    event_url: Optional[str]
    event_async_url: Optional[str]


def update_web_server_info(**kwargs: Unpack[TWebServerInfo]):
    if App._web_server_info is None:
        App._web_server_info = {}

    data = {k: v for k, v in kwargs.items() if v is not None}

    App._web_server_info.update(data)



import typing
from instaui.runtime import ui_state_scope


_T = typing.TypeVar("_T")


def injection_key(key: str) -> _T:
    """creates a key for injection

    Args:
        key (str): key to use for injection

    Examples:
    .. code-block:: python
        text_key :str = injection_key("text")

        # state model
        class Person(ui.StateModel):
            name:str

        person_key :Person = injection_key(Person(name='foo'))
    """
    return key  # type: ignore


def inject_state(key: _T) -> _T:
    return typing.cast(_T, ui_state_scope.load_state(key))  # type: ignore


def provide_state(key: _T, obj: _T) -> _T:
    ui_state_scope.save_state(key, obj)
    return obj



from instaui.common.jsonable import Jsonable
from instaui.vars._types import InputBindingType
from instaui.vars.mixin_types.py_binding import CanInputMixin
from instaui.runtime._app import get_app_slot, get_current_scope


class JsFn(Jsonable, CanInputMixin):
    """
    Creates a JavaScript function object from a raw code string.
    Valid targets include: `js_computed`, `js_watch`, and similar JS-bound methods.

    Args:
        code (str): Valid JavaScript function definition string.

    Example:
    .. code-block:: python
        a = ui.state(1)
        add = ui.js_fn(code="(a,b)=> a+b ")
        result = ui.js_computed(inputs=[add, a], code="(add,a)=>  add(a,10) ")

        html.number(a)
        ui.label(result)
    """

    def __init__(self, code: str, *, execute_immediately=False, global_scope=False):
        self.code = code

        scope = get_app_slot().top_scope if global_scope else get_current_scope()
        self.__register_info = scope.register_js_fn_task(self)

        self._execute_immediately = execute_immediately

    def _to_input_config(self):
        return {
            "sid": self.__register_info.scope_id,
            "id": self.__register_info.var_id,
        }

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["id"] = self.__register_info.var_id

        if self._execute_immediately is True:
            data["immediately"] = 1

        return data

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.JsFn



from instaui.components.element import Element
from instaui.vars.types import TMaybeRef


def teleport(
    to: TMaybeRef[str],
    *,
    defer: bool = True,
    disabled: TMaybeRef[bool] = False,
):
    """
    Creates a teleport element that can be used to move a component to a different part of the DOM.


    """
    ele = Element(tag="teleport")
    ele.props({"to": to})

    if defer is not True:
        ele.props({"defer": False})

    if disabled is not False:
        ele.props({"disabled": disabled})

    return ele



from __future__ import annotations
from typing import (
    Dict,
    List,
    Optional,
    Union,
)
from instaui.components.value_element import ValueElement
from instaui.components.element import Element
from instaui.event.event_mixin import EventMixin
from instaui.vars.types import TMaybeRef
from instaui.components.vfor import VFor
from instaui.components.mixins import CanDisabledMixin

_T_Select_Value = Union[List[str], str]


class Select(CanDisabledMixin, ValueElement[Union[List[str], str]]):
    def __init__(
        self,
        value: Union[_T_Select_Value, TMaybeRef[_T_Select_Value], None] = None,
        *,
        model_value: Union[str, TMaybeRef[str], None] = None,
    ):
        super().__init__("select", value, is_html_component=True)

        if model_value is not None:
            self.props({"value": model_value})

    def on_change(
        self,
        handler: EventMixin,
        *,
        extends: Optional[List] = None,
    ):
        self.on("change", handler)
        return self

    @classmethod
    def from_list(
        cls,
        options: TMaybeRef[List],
        value: Union[_T_Select_Value, TMaybeRef[_T_Select_Value], None] = None,
        *,
        model_value: Union[str, TMaybeRef[str], None] = None,
    ) -> Select:
        with cls(value, model_value=model_value) as select:
            with VFor(options) as item:
                Select.Option(item)  # type: ignore

        return select

    @classmethod
    def from_records(
        cls,
        options: TMaybeRef[List[Dict]],
        value: Union[_T_Select_Value, TMaybeRef[_T_Select_Value], None] = None,
        *,
        model_value: Union[str, TMaybeRef[str], None] = None,
    ) -> Select:
        with cls(value, model_value=model_value) as select:
            with VFor(options) as item:
                Select.Option(item["text"], item["value"])  # type: ignore

        return select

    class Option(Element, CanDisabledMixin):
        def __init__(
            self,
            text: Optional[TMaybeRef[str]] = None,
            value: Optional[TMaybeRef[str]] = None,
            disabled: Optional[TMaybeRef[bool]] = None,
        ):
            props = {
                key: value
                for key, value in {
                    "text": text,
                    "value": value,
                    "disabled": disabled,
                }.items()
                if value is not None
            }
            super().__init__("option")

            self._props.update(props)



from __future__ import annotations
from typing import Any, Dict, Optional
from instaui import consts
from ._app import get_app_slot


class Context:
    _instance: Optional[Context] = None

    @classmethod
    def get_instance(cls, *args, **kwargs):
        if cls._instance is None:
            cls._instance = cls(*args, **kwargs)
        return cls._instance

    def __init__(self):
        self._app_mode: consts._T_App_Mode = "web"
        self._debug_mode: bool = False
        self._page_params: Dict[str, Any] = {}

    @property
    def app_mode(self):
        return self._app_mode

    @property
    def debug_mode(self):
        return self._debug_mode

    @property
    def page_path(self):
        return get_app_slot().page_path

    @property
    def page_params(self):
        return get_app_slot().page_params

    @property
    def query_params(self):
        return get_app_slot().query_params


def get_context():
    return Context.get_instance()


def set_debug(debug=True):
    get_context()._debug_mode = debug



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Union
from instaui.components.element import Element

if TYPE_CHECKING:
    from instaui.vars.types import TMaybeRef


class Li(Element):
    def __init__(
        self,
        text: Union[Any, TMaybeRef[Any], None] = None,
    ):
        super().__init__("li")

        if text:
            self.props({"innerText": text})



def show_error(msg: str):
    def error_handler(error: Exception):
        from instaui.components.html.paragraph import Paragraph

        Paragraph(msg).style(
            "border: 1px dashed red;padding: 1em;font-weight: bold;font-size: medium;"
        )

    return error_handler



from __future__ import annotations
import typing

from instaui.common.jsonable import Jsonable
from instaui.vars.path_var import PathVar

from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.mixin_types.pathable import CanPathPropMixin


class RouteParamsVar(
    Jsonable,
    PathVar,
    ObservableMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
):
    VAR_TYPE = "rp"

    def __init__(
        self,
        prop: typing.Literal["params", "path", "fullPath"] = "params",
    ) -> None:
        super().__init__()
        self._prop = prop

    def __to_binding_config(self):
        return {"type": "rp", "prop": self._prop}

    def _to_pathable_binding_config(self) -> typing.Dict:
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> typing.Dict:
        return self.__to_binding_config()

    def _to_observable_config(self):
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_json_dict(self):
        raise NotImplementedError("RouteParamsVar is not json serializable")



from typing import Callable
from instaui.launch_collector import get_launch_collector


def on_page_request_lifespan(fn: Callable):
    """Register a function to be called on each page request.

    Args:
        fn (Callable): A function to be called on each page request.

    Examples:
    .. code-block:: python
        @ui.on_page_request_lifespan
        def _():
            print("page request start")
            yield
            print("page request end")

        # can stop the lifespan by calling the returned function
        stop_lifespan = ui.on_page_request_lifespan(lambda: None)
        stop_lifespan()
    """

    remove_key = get_launch_collector().add_page_request_lifespan(fn)
    return lambda: get_launch_collector().remove_page_request_lifespan(remove_key)



from __future__ import annotations
from dataclasses import dataclass, field
from pathlib import Path
from typing import List, Optional
from instaui.runtime._app import get_app_slot


@dataclass(frozen=True)
class PluginDependencyInfo:
    name: str = field(hash=True)
    esm: Path = field(hash=False)
    externals: Optional[List[Path]] = field(
        default_factory=list, compare=False, hash=False
    )
    css: List[Path] = field(default_factory=list, compare=False, hash=False)


def register_plugin(
    name: str,
    esm: Path,
    *,
    externals: Optional[List[Path]] = None,
    css: Optional[List[Path]] = None,
):
    info = PluginDependencyInfo(f"plugin/{name}", esm, externals or [], css or [])

    get_app_slot().use_plugin_dependency(info)
    return info



import typing
from instaui.vars._types import InputBindingType
from instaui.vars.mixin_types.py_binding import CanInputMixin


class InputSilentData(CanInputMixin):
    def __init__(self, value: typing.Union[CanInputMixin, typing.Any]) -> None:
        self.value = value

    def is_const_value(self) -> bool:
        return not isinstance(self.value, CanInputMixin)

    def _to_input_config(self):
        if isinstance(self.value, CanInputMixin):
            return self.value._to_input_config()
        else:
            return self.value

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref



from __future__ import annotations
from typing import Any, Dict, Generic, Tuple, TypeVar, Union, cast

from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_current_scope
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.vars.mixin_types.pathable import CanPathPropMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.path_var import PathVar


_T = TypeVar("_T")


class VForItemProxy(
    PathVar,
    CanInputMixin,
    ObservableMixin,
    CanOutputMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
    Jsonable,
    Generic[_T],
):
    def __init__(self, vfor_item: VForItem[_T]):
        self._vfor_item = vfor_item

    def __getattr__(self, name: str):
        return self._vfor_item[name]

    @property
    def __vfor_item_(self):
        return super().__getattribute__("_vfor_item")

    def __getitem__(self, name):
        return self.__vfor_item_[name]

    def _to_element_binding_config(self) -> Dict:
        return self.__vfor_item_._to_element_binding_config()

    def _to_input_config(self):
        return self.__vfor_item_._to_input_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__vfor_item_._to_path_prop_binding_config()

    def _to_output_config(self):
        return self.__vfor_item_._to_output_config()

    def _to_str_format_binding(self, order: int) -> Tuple[str, str]:
        return self.__vfor_item_._to_str_format_binding(order)

    def _to_pathable_binding_config(self) -> Dict:
        return self.__vfor_item_._to_pathable_binding_config()

    def _to_observable_config(self):
        return self.__vfor_item_._to_observable_config()

    def _to_json_dict(self):
        return self.__vfor_item_._to_json_dict()

    def _to_event_output_type(self) -> OutputSetType:
        return self.__vfor_item_._to_event_output_type()

    def _to_event_input_type(self) -> InputBindingType:
        return self.__vfor_item_._to_event_input_type()


class VForItem(
    PathVar,
    CanInputMixin,
    ObservableMixin,
    CanOutputMixin,
    CanPathPropMixin,
    ElementBindingMixin[_T],
    StrFormatBindingMixin,
    Jsonable,
    Generic[_T],
):
    SCOPE_TYPE = "fv"

    def __init__(self):
        super().__init__()
        scope = get_current_scope()
        self.__register_info = scope.register_vfor_item_task(self)

    def __getattr__(self, name: str):
        return self[name]

    # def __getitem__(self, name):
    #     return self[name]

    @property
    def dict_key(self):
        return cast(Any, VForItemKey())

    @property
    def dict_value(self):
        return self

    @property
    def index(self) -> int:
        return cast(int, VForIndex())

    @property
    def value(self) -> _T:
        return cast(_T, self)

    # @property
    # def proxy(self):
    #     return cast(_T, VForItemProxy(self))

    def _to_binding_config(self) -> Union[Jsonable, Dict]:
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_output_config(self):
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_pathable_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_observable_config(self):
        return self.__to_binding_config()

    def __to_binding_config(self):
        data: Dict = {
            "sid": self.__register_info.scope_id,
            "id": self.__register_info.var_id,
        }

        return data

    def _to_json_dict(self):
        data: Dict = {
            "id": self.__register_info.var_id,
        }

        return data

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.Ref

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref


class VForIndex(
    CanInputMixin,
    CanPathPropMixin,
    ElementBindingMixin,
    StrFormatBindingMixin,
    Jsonable,
):
    SCOPE_TYPE = "fi"

    def __init__(self):
        super().__init__()
        scope = get_current_scope()
        self.__register_info = scope.register_vfor_index_task(self)

    def __to_binding_config(self):
        data: Dict = {
            "sid": self.__register_info.scope_id,
            "id": self.__register_info.var_id,
        }

        return data

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_json_dict(self):
        return {
            "id": self.__register_info.var_id,
        }

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref


class VForItemKey(
    CanInputMixin,
    CanPathPropMixin,
    ElementBindingMixin,
    StrFormatBindingMixin,
    Jsonable,
):
    SCOPE_TYPE = "fk"

    def __init__(self):
        super().__init__()
        scope = get_current_scope()
        self.__register_info = scope.register_vfor_key_task(self)

    def __to_binding_config(self):
        data: Dict = {
            "sid": self.__register_info.scope_id,
            "id": self.__register_info.var_id,
        }

        return data

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_json_dict(self):
        return {
            "id": self.__register_info.var_id,
        }

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref


# class VForDict(
#     CanInputMixin,
#     CanOutputMixin,
#     StrFormatBindingMixin,
#     ElementBindingMixin,
#     Jsonable,
# ):
#     def __init__(self, vfor: VFor):
#         self._vfor = vfor

#     @property
#     def dict_key(self):
#         return self._vfor.current[0]

#     @property
#     def dict_value(self):
#         return self._vfor.current[1]

#     @contextmanager
#     def with_index(self):
#         self.__enter__()
#         yield self, cast(int, VForIndex(self._vfor))

#     def __enter__(self):
#         self._vfor.__enter__()
#         return self

#     def __exit__(self, *_) -> None:
#         return self._vfor.__exit__(*_)

#     def _to_element_binding_config(self) -> Dict:
#         return self.dict_value._to_element_binding_config()

#     def _to_input_config(self):
#         return self.dict_value._to_input_config()

#     def _to_output_config(self):
#         return self.dict_value._to_output_config()

#     def _to_json_dict(self):
#         return self.dict_value._to_json_dict()

#     def _to_event_output_type(self) -> EventOutputType:
#         raise ValueError("VForDict does not support output events")

#     def _to_event_input_type(self) -> EventInputType:
#         return EventInputType.Ref


# class VForWithIndex(Generic[_T]):
#     def __init__(self, vfor: VFor[_T]):
#         self._vfor = vfor

#     def __enter__(self):
#         self._vfor.__enter__()
#         return cast(_T, self._vfor.current.proxy), cast(int, VForIndex(self._vfor))

#     def __exit__(self, *_) -> None:
#         return self._vfor.__exit__(*_)


TVForItem = VForItem
TVForIndex = VForIndex



from typing import Tuple
from abc import ABC


class StrFormatBindingMixin(ABC):
    def _to_str_format_binding(self, order: int) -> Tuple[str, str]:
        var_name = f"___ref_var{order}"
        return var_name, f"${{__Vue.toValue({var_name})}}"



from __future__ import annotations
from typing import Any, Dict, Protocol, TYPE_CHECKING, Union
from typing_extensions import Self
from instaui.vars.types import TMaybeRef

if TYPE_CHECKING:
    from instaui.components.element import Element


class PropsProtocol(Protocol):
    def props(self, add: Union[str, Dict[str, Any], TMaybeRef]) -> Self: ...


class CanDisabledMixin:
    def disabled(self: PropsProtocol, disabled: TMaybeRef[bool] = True) -> Element:
        return self.props({"disabled": disabled})  # type: ignore



from typing import Dict, Generic, TypeVar, Protocol
from abc import ABC, abstractmethod

T = TypeVar("T")


class ElementBindingMixin(ABC, Generic[T]):
    @abstractmethod
    def _to_element_binding_config(self) -> Dict:
        pass


class ElementBindingProtocol(Protocol):
    def _to_element_binding_config(self) -> Dict: ...



from __future__ import annotations
import abc
from typing import TYPE_CHECKING, List, Optional
from instaui.event.event_mixin import EventMixin


if TYPE_CHECKING:
    from instaui.components.element import Element


class InputEventMixin:
    @abc.abstractmethod
    def _input_event_mixin_element(self) -> Element:
        pass

    def on_change(
        self,
        handler: EventMixin,
        *,
        extends: Optional[List] = None,
        key: Optional[str] = None,
    ):
        self._input_event_mixin_element().on("change", handler, extends=extends)
        return self

    def on_input(
        self,
        handler: EventMixin,
        *,
        extends: Optional[List] = None,
        key: Optional[str] = None,
    ):
        self._input_event_mixin_element().on("input", handler, extends=extends)
        return self



from pathlib import Path
from instaui.runtime._app import get_app_slot


_STATIC_DIR = Path(__file__).parent / "static"
_CSS_FILE = _STATIC_DIR / "shadcn-classless.css"


def use_shadcn_classless(value=True):
    """Use shadcn-classless.css or not.

    Args:
        value (bool, optional): Whether to use shadcn-classless.css or not. Defaults to True.

    Example:

    App default use shadcn-classless.css:
    .. code-block:: python

        @ui.page('/')
        def index():

            # Create a container with shadcn-classless class
            with html.div().classes("shadcn-classless"):
                html.h1("Hello, world!")

                with html.form():
                    html.input()
                    html.button("Submit")

    Can be disabled:
    .. code-block:: python
        # App does not use shadcn-classless
        ui.use_shadcn_classless(False)

    """

    if value:
        get_app_slot()._html_resource.add_css_link(_CSS_FILE)

    else:
        get_app_slot()._html_resource.remove_css_link(_CSS_FILE)



from __future__ import annotations
from typing import List, Optional
from instaui.components.element import Element
from instaui.vars.types import TMaybeRef
from instaui.event.event_mixin import EventMixin


class Timer(Element, esm="./timer.js"):
    """
    A timer component that triggers events at specified intervals.

    Args:
        interval_seconds (float): Time in seconds between repeated events.
        active (Optional[TMaybeRef[bool]], optional): If True, starts the timer immediately.
            Accepts reactive values (e.g., state references) for dynamic control. Defaults to None (equivalent to True).
        immediate (Optional[bool], optional): If True, triggers the first event immediately. Defaults to None (equivalent to True).

    Example:
    .. code-block:: python
        from instaui import ui, html

        @ui.page('/')
        def index():
            x = ui.state(0)
            active = ui.state(True)

            @ui.event(inputs=[x], outputs=[x])
            def on_tick(x):
                return x + 1

            ui.timer(1, active=active).on_tick(on_tick)

            html.checkbox(active)
            html.span(x)
    """

    def __init__(
        self,
        interval_seconds: float,
        *,
        active: Optional[TMaybeRef[bool]] = None,
        immediate: Optional[bool] = None,
    ):
        super().__init__("template")

        self.props(
            {
                "intervalSeconds": interval_seconds,
                "active": active,
                "immediate": immediate,
            }
        )

    def on_tick(
        self,
        handler: EventMixin,
        *,
        extends: Optional[List] = None,
    ):
        """
        Registers an event handler for the "tick" event.
        """
        return self.on("tick", handler, extends=extends)

    def on_stop(
        self,
        handler: EventMixin,
        *,
        extends: Optional[List] = None,
    ):
        """
        Registers an event handler for the "stop" event.
        """

        return self.on("stop", handler, extends=extends)

    @classmethod
    def once(cls, delay_seconds: float):
        """
        Creates a timer that triggers only once.

        Args:
            delay_seconds (float): Time in seconds before the timer triggers.

        Example:
        .. code-block:: python
            msg = ui.state('')
            on_done = ui.js_event(outputs=[msg],code="()=> 'Done!'")

            ui.timer.once(1).on_tick(on_done)
            html.span(msg) # will show "Done!" after 1 second

        """
        return cls(delay_seconds, immediate=False).props({"once": True})



from __future__ import annotations
from pathlib import Path
from typing import Any, Dict, Literal, Optional, Union
from instaui.common.jsonable import dumps, dumps2dict
from instaui.runtime._app import get_app_slot
from instaui.tailwind._index import use_tailwind

use_tailwind = use_tailwind


def add_css_link(href: Union[str, Path]):
    """Add a link to a CSS file to the HTML document.

    Args:
        href (Union[str, Path]): The path to the CSS file.
    """
    get_app_slot()._html_resource.add_css_link(href)


def remove_css_link(href: Union[str, Path]):
    """Remove a link to a CSS file from the HTML document.

    Args:
        href (Union[str, Path]): The path to the CSS file.
    """
    get_app_slot()._html_resource.remove_css_link(href)


def add_js_link(
    link: Union[str, Path],
    *,
    type: Optional[Literal["module"]] = None,
):
    """Add a link to a JavaScript file to the HTML document.

    Args:
        link (Union[str, Path]): The path to the JavaScript file.
        type (Optional[Literal[&quot;module&quot;]], optional): The type of the JavaScript file. Defaults to None.
    """

    attrs = {
        "type": type,
    }

    get_app_slot()._html_resource.add_js_link(link, attrs=attrs)


add_style = get_app_slot()._html_resource.add_style_tag


def use_page_title(title: str):
    """Set the title of the HTML document.

    Args:
        title (str): The title of the HTML document.
    """
    get_app_slot()._html_resource.title = title


def use_favicon(favicon: Path):
    """Set the favicon of the HTML document.

    Args:
        favicon (Path): The path to the favicon.
    """
    get_app_slot()._html_resource.favicon = favicon


def add_js_code(code: str, *, script_attrs: Optional[Dict[str, Any]] = None):
    """Add a script tag to the HTML document with the given JavaScript code.

    Args:
        code (str): The JavaScript code.
        script_attrs (Optional[Dict[str, Any]], optional): The attributes of the script tag. Defaults to None.
    """
    get_app_slot()._html_resource.add_script_tag(code, script_attrs=script_attrs)


def add_vue_app_use(name: str):
    get_app_slot()._html_resource.add_vue_app_use(name)


def to_config_data() -> Dict:
    app = get_app_slot()
    data = dumps2dict(app)

    # Exclude scopes that contain only the id field.
    scopes = [s for s in dumps2dict(app._scopes) if len(s) > 1]
    data["scopes"] = scopes
    return data


def to_json(indent=False):
    return dumps(to_config_data(), indent=indent)



from __future__ import annotations
import inspect
import typing
from typing_extensions import ParamSpec
from . import _types
from . import _utils

from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_app_slot, get_current_scope
from instaui.handlers import watch_handler

from instaui.vars.mixin_types.py_binding import (
    CanOutputMixin,
    outputs_to_config,
    _assert_outputs_be_can_output_mixin,
)
from instaui.vars.mixin_types.common_type import TObservableInput
from instaui._helper import observable_helper
from instaui import pre_setup as _pre_setup

_SYNC_TYPE = "sync"
_ASYNC_TYPE = "async"

P = ParamSpec("P")
R = typing.TypeVar("R")


class WebWatch(Jsonable, typing.Generic[P, R]):
    def __init__(
        self,
        func: typing.Callable[P, R],
        inputs: typing.Optional[typing.Sequence[TObservableInput]] = None,
        outputs: typing.Optional[typing.Sequence[CanOutputMixin]] = None,
        immediate: bool = True,
        deep: typing.Union[bool, int] = True,
        once: bool = False,
        flush: typing.Optional[_types.TFlush] = None,
        pre_setup: typing.Optional[typing.Dict] = None,
        _debug: typing.Optional[typing.Any] = None,
    ) -> None:
        if pre_setup:
            _pre_setup._check_args(pre_setup)

        outputs = [] if outputs is None else outputs

        _assert_outputs_be_can_output_mixin(outputs)

        inputs = observable_helper.auto_made_inputs_to_slient(inputs, outputs)

        self._inputs, self._is_slient_inputs, self._is_data = (
            observable_helper.analyze_observable_inputs(list(inputs or []))
        )

        self._outputs = outputs
        self._fn = func
        self._immediate = immediate
        self._deep = deep
        self._once = once
        self._flush = flush
        self._debug = _debug
        self._pre_setup = pre_setup

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self._fn(*args, **kwargs)

    def _to_json_dict(self):
        data = super()._to_json_dict()

        app = get_app_slot()

        if app.mode == "web":
            hkey = watch_handler.create_handler_key(
                page_path=app.page_path,
                handler=self._fn,
            )

            watch_handler.register_handler(hkey, self._fn, len(self._outputs))

            data["fType"] = (
                _ASYNC_TYPE if inspect.iscoroutinefunction(self._fn) else _SYNC_TYPE
            )
            data["key"] = hkey
            if self._inputs:
                data["inputs"] = self._inputs

            if sum(self._is_slient_inputs) > 0:
                data["slient"] = self._is_slient_inputs

            if sum(self._is_data) > 0:
                data["data"] = self._is_data

            if self._debug:
                data["debug"] = self._debug

            if self._outputs:
                data["outputs"] = outputs_to_config(self._outputs)

            if self._immediate is not True:
                data["immediate"] = self._immediate

            if self._deep is not True:
                _utils.assert_deep(self._deep)
                data["deep"] = self._deep
            if self._once is not False:
                data["once"] = self._once
            if self._flush is not None:
                data["flush"] = self._flush
            if self._pre_setup:
                data["preSetup"] = _pre_setup.convert_config(self._pre_setup)

            return data

        return {}


def watch(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    outputs: typing.Optional[typing.Sequence] = None,
    immediate: bool = True,
    deep: typing.Union[bool, int] = True,
    once: bool = False,
    flush: typing.Optional[_types.TFlush] = None,
    pre_setup: typing.Optional[typing.Dict] = None,
    _debug: typing.Optional[typing.Any] = None,
):
    """
    Creates an observer that tracks changes in reactive sources and triggers callbacks.

    Args:
        inputs (typing.Optional[typing.Sequence], optional): Reactive sources to observe (state objects or computed properties).
                                   Changes to these sources trigger the watcher callback.
        outputs (typing.Optional[typing.Sequence], optional): Output targets associated with this watcher.
                                    Used for coordination with computed properties or other observers.
        immediate (bool, optional): If True, executes callback immediately after creation with current values. Defaults to True.
        deep (typing.Union[bool, int], optional): Controls depth of change detection:
                               - True: Recursively tracks nested properties
                               - False: Shallow comparison only
                               - int: Maximum depth level to track (for complex objects).
                               Defaults to True.
        once (bool, optional):  If True, automatically stops observation after first trigger. Defaults to False.
        flush (typing.Optional[_types.TFlush], optional): Controls when to flush updates:
                                      - 'sync': Execute immediately on change
                                      - 'post': Batch updates and execute after current tick
                                      - 'pre': Execute before render phase (if applicable)
        pre_setup (typing.Optional[typing.Dict], optional): A dictionary of pre-setup actions to be executed before the event executes.


    # Example:
    .. code-block:: python
        from instaui import ui, html

        num = ui.state(0)
        msg = ui.state('')

        @ui.watch(inputs=[num], outputs=[msg])
        def when_num_change(num):
            return f"The number is {num}"

        html.number(num)
        ui.label(msg)

    list append:
    .. code-block:: python
        from instaui import ui, html

        num = ui.state(0)
        msg = ui.state([])

        @ui.watch(inputs=[num, msg], outputs=[msg])
        def when_num_change(num, msg:list):
            msg.append(f"The number changed to {num}")
            return msg

        html.number(num)
        ui.label(msg)

    """

    def wrapper(func: typing.Callable[P, R]):
        obj = WebWatch(
            func,
            inputs,
            outputs=outputs,
            immediate=immediate,
            deep=deep,
            once=once,
            flush=flush,
            pre_setup=pre_setup,
            _debug=_debug,
        )

        get_current_scope().register_web_watch(obj)

    return wrapper



from typing import List, Optional
from instaui.common.jsonable import Jsonable
from instaui.vars._types import OutputSetType
from instaui.vars.mixin_types.py_binding import CanOutputMixin
from pydantic import BaseModel


class RouterOutput(Jsonable, CanOutputMixin):
    def __init__(self):
        self.type = "routeAct"

    def _to_output_config(self):
        return {}

    def _to_json_dict(self):
        data = super()._to_json_dict()

        return data

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.RouterAction


class RouterMethod(BaseModel):
    fn: str
    args: Optional[List]



from __future__ import annotations
from typing import Any, Dict, Mapping, Optional, Union

from instaui.common.jsonable import Jsonable

from instaui.runtime._app import get_current_scope
from instaui.vars._types import InputBindingType
from instaui.vars.path_var import PathVar
from instaui.vars.mixin_types.var_type import VarMixin
from instaui.vars.mixin_types.element_binding import (
    ElementBindingMixin,
    ElementBindingProtocol,
)
from instaui.vars.mixin_types.py_binding import CanInputMixin
from instaui.vars.mixin_types.pathable import CanPathPropMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.mixin_types.observable import ObservableMixin


class VueComputed(
    Jsonable,
    PathVar,
    VarMixin,
    CanInputMixin,
    ObservableMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin,
):
    def __init__(
        self,
        fn_code: str,
        bindings: Optional[Mapping[str, Union[ElementBindingProtocol, Any]]] = None,
    ) -> None:
        self.code = fn_code
        self._bindings = bindings
        scope = get_current_scope()
        self.__register_info = scope.register_vue_computed_task(self)

        self._sid = scope.id
        self._id = scope.generate_vars_id()

        if bindings:
            const_bind = []
            self.bind = {}

            for k, v in bindings.items():
                is_binding = isinstance(v, ElementBindingMixin)
                self.bind[k] = v._to_element_binding_config() if is_binding else v
                const_bind.append(int(not is_binding))

            if any(i == 1 for i in const_bind):
                self.const = const_bind

    def __to_binding_config(self):
        return {
            "id": self.__register_info.var_id,
            "sid": self.__register_info.scope_id,
        }

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_pathable_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_observable_config(self):
        return self.__to_binding_config()

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["id"] = self.__register_info.var_id
        return data

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref


TVueComputed = VueComputed



import typing
from instaui.vars.mixin_types.py_binding import (
    CanInputMixin,
    CanOutputMixin,
    inputs_to_config,
    outputs_to_config,
)
from instaui.common.jsonable import Jsonable
from .event_mixin import EventMixin


class JsEvent(Jsonable, EventMixin):
    def __init__(
        self,
        code: str,
        inputs: typing.Optional[typing.Sequence[CanInputMixin]] = None,
        outputs: typing.Optional[typing.Sequence[CanOutputMixin]] = None,
    ):
        self._is_const_data = [
            int(not isinstance(input, CanInputMixin)) for input in inputs or []
        ]
        self._org_inputs = list(inputs or [])
        self._org_outputs = list(outputs or [])
        self._inputs = inputs or []
        self._outputs = outputs or []
        self.code = code

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["type"] = self.event_type()

        if self._inputs:
            data["inputs"] = inputs_to_config(self._inputs)

        if self._outputs:
            data["sets"] = outputs_to_config(self._outputs)

        if sum(self._is_const_data) > 0:
            data["data"] = self._is_const_data

        return data

    def copy_with_extends(self, extends: typing.Sequence):
        return js_event(
            code=self.code,
            inputs=self._org_inputs + list(extends),
            outputs=self._org_outputs,
        )

    def event_type(self):
        return "js"


def js_event(
    *,
    inputs: typing.Optional[typing.Sequence] = None,
    outputs: typing.Optional[typing.Sequence] = None,
    code: str,
):
    """
    Creates a client-side event handler decorator for binding JavaScript logic to UI component events.

    Args:
        inputs (typing.Optional[typing.Sequence], optional):Reactive sources (state variables, computed values)
                                   that should be passed to the event handler. These values
                                   will be available in the JavaScript context through the `args` array.
        outputs (typing.Optional[typing.Sequence], optional): Targets (state variables, UI elements) that should
                                    update when this handler executes. Used for coordinating
                                    interface updates after the event is processed.
        code (str): JavaScript code to execute when the event is triggered.

    # Example:
    .. code-block:: python
        from instaui import ui, html

        a = ui.state(0)

        plus_one = ui.js_event(inputs=[a], outputs=[a], code="a =>a + 1")

        html.button("click me").on_click(plus_one)
        html.paragraph(a)

    """
    return JsEvent(inputs=inputs, outputs=outputs, code=code)



from __future__ import annotations
from abc import abstractmethod
from typing import List, Optional, Union
from typing_extensions import Self
from instaui.vars._types import InputBindingType, OutputSetType
from instaui.vars.mixin_types.pathable import PathableMixin, CanPathPropMixin
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.py_binding import CanInputMixin, CanOutputMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin


class PathVar(PathableMixin):
    def __getitem__(self, item: Union[str, int, CanPathPropMixin]):
        return PathTrackerBindable(self)[item]

    def not_(self):
        return PathTrackerBindable(self).not_()

    def __add__(self, other: str):
        return PathTrackerBindable(self) + other

    def __radd__(self, other: str):
        return other + PathTrackerBindable(self)


class PathTracker:
    def __init__(self, paths: Optional[List[Union[str, List[str]]]] = None):
        self.paths = paths or []

    def __getitem__(self, key) -> Self:
        return self.__new_self__([*self.paths, key])

    def __getattr__(self, key) -> Self:
        return self.__new_self__([*self.paths, key])

    def not_(self) -> Self:
        return self.__new_self__([*self.paths, ["!"]])

    def __add__(self, other: str) -> Self:
        return self.__new_self__([*self.paths, ["+", other]])

    def __radd__(self, other: str) -> Self:
        return self.__new_self__([*self.paths, ["~+", other]])

    @abstractmethod
    def __new_self__(self, paths: List[Union[str, List[str]]]) -> Self:
        pass


class PathTrackerBindable(
    PathTracker,
    CanInputMixin,
    ObservableMixin,
    CanOutputMixin,
    ElementBindingMixin,
    StrFormatBindingMixin,
):
    def __init__(self, source: PathableMixin):
        super().__init__()
        self.__source = source

    def __new_self__(self, paths: List[Union[str, List[str]]]) -> PathTrackerBindable:
        obj = PathTrackerBindable(self.__source)
        obj.paths = paths
        return obj

    def _to_element_binding_config(self):
        return self._to_json_dict()

    def _to_input_config(self):
        return self._to_json_dict()

    def _to_output_config(self):
        return self._to_json_dict()

    def _to_observable_config(self):
        return self._to_json_dict()

    def _to_json_dict(self):
        data = self.__source._to_pathable_binding_config()

        if self.paths:
            data["path"] = [
                ["bind", path._to_path_prop_binding_config()]
                if isinstance(path, CanPathPropMixin)
                else path
                for path in self.paths
            ]

        return data

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref

    def _to_event_output_type(self) -> OutputSetType:
        return OutputSetType.Ref



from __future__ import annotations
import typing
from dataclasses import dataclass, field
from .env import env
from instaui.common.jsonable import dumps
from instaui.runtime.context import get_context
from instaui.runtime.dataclass import JsLink, VueAppComponent


if typing.TYPE_CHECKING:
    from instaui.runtime.resource import StyleTag


_html_template = env.get_template("web.html")


@dataclass
class WebTemplateModel:
    version: str
    vue_js_link: str
    instaui_js_link: str
    config_dict: typing.Dict[str, typing.Any] = field(default_factory=dict)
    extra_import_maps: typing.Dict[str, str] = field(default_factory=dict)
    css_links: typing.List[str] = field(default_factory=list)
    style_tags: typing.List[StyleTag] = field(default_factory=list)
    js_links: typing.List[JsLink] = field(default_factory=list)
    script_tags: typing.List[str] = field(default_factory=list)
    vue_app_use: typing.List[str] = field(default_factory=list)
    vue_app_component: typing.List[VueAppComponent] = field(default_factory=list)
    prefix: str = ""
    title: typing.Optional[str] = None
    favicon_url: str = ""

    def add_extra_import_map(self, name: str, url: str):
        self.extra_import_maps[name] = url

    @property
    def import_maps_string(self):
        data = {
            **self.extra_import_maps,
            "vue": self.vue_js_link,
            "instaui": self.instaui_js_link,
        }

        return dumps(data)

    @property
    def is_debug(self):
        context = get_context()
        return context.debug_mode


def render_web_html(model: WebTemplateModel):
    return _html_template.render(model=model)



from instaui.runtime.context import get_context


def check_web_only_mode_or_error(method_name: str = ""):
    mode = get_context().app_mode
    if mode != "web" and mode != "webview":
        raise Exception(
            f"{method_name} This is a web-only feature. Please use the 'web' or 'webview' mode."
        )



from typing import cast
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.vue_computed import VueComputed


def str_format(template: str, *args, **kwargs):
    bindings = {}

    tran_args = []

    for idx, arg in enumerate(args):
        is_mixin = isinstance(arg, StrFormatBindingMixin)
        value = (
            cast(StrFormatBindingMixin, arg)._to_str_format_binding(idx)
            if is_mixin
            else arg
        )
        tran_args.append(value[-1] if is_mixin else value)
        if is_mixin:
            bindings[value[0]] = arg

    tran_kwargs = {}

    for idx, (k, v) in enumerate(kwargs.items()):
        is_mixin = isinstance(v, StrFormatBindingMixin)
        value = (
            cast(StrFormatBindingMixin, v)._to_str_format_binding(idx)
            if is_mixin
            else v
        )
        tran_kwargs[k] = value[-1] if is_mixin else value
        if is_mixin:
            bindings[value[0]] = v

    code = "()=>`" + template.format(*tran_args, **tran_kwargs) + "`"
    return VueComputed(code, bindings=bindings)



from __future__ import annotations
from typing import (
    TYPE_CHECKING,
    List,
    Optional,
)
from instaui.components.element import Element
from instaui.event.event_mixin import EventMixin
from instaui.vars.types import TMaybeRef
from instaui.components.mixins import CanDisabledMixin

if TYPE_CHECKING:
    pass


class Button(Element, CanDisabledMixin):
    def __init__(
        self,
        text: Optional[TMaybeRef[str]] = None,
    ):
        super().__init__("button")

        if text is not None:
            self.props(
                {
                    "innerText": text,
                }
            )

    def on_click(
        self,
        handler: EventMixin,
        *,
        extends: Optional[List] = None,
    ):
        self.on("click", handler, extends=extends)
        return self



from typing import Any, Dict, List, Optional, Sequence
from instaui.runtime._app import get_app_slot
from instaui.response import response_data
import pydantic


class ResponseData(pydantic.BaseModel):
    values: Optional[List[Any]] = None
    types: Optional[Sequence[int]] = None


def update_app_page_info(data: Dict):
    app = get_app_slot()

    page_info = data.get("page", {})
    app._page_path = page_info["path"]

    if "params" in page_info:
        app._page_params = page_info["params"]

    if "queryParams" in page_info:
        app._query_params = page_info["queryParams"]


def response_web_data(outputs_binding_count: int, result: Any):
    return ResponseData(**response_data(outputs_binding_count, result))



from __future__ import annotations
from typing import TYPE_CHECKING, Any, Optional, Union
from instaui.components.element import Element

if TYPE_CHECKING:
    from instaui.vars.types import TMaybeRef


class Label(Element):
    def __init__(
        self,
        text: Union[Any, TMaybeRef[Any], None] = None,
        *,
        for_: Optional[TMaybeRef[str]] = None,
    ):
        super().__init__("label")

        self.props(
            {
                "innerText": text,
                "for": for_,
            }
        )



import typing
from instaui.common.jsonable import Jsonable
from instaui.vars.mixin_types.py_binding import CanInputMixin
from instaui.vars._types import InputBindingType
from .event_mixin import EventMixin


class VueEvent(Jsonable, EventMixin):
    """
    Create an event object that can be bound to a UI component's event listener.

    This function generates a callable event handler with optional contextual bindings.
    The event logic is defined via a code string, which can reference bound variables.

    Args:
        code (str): A string containing the executable logic for the event handler.
                    Typically contains a function body or expression that utilizes bound variables.
        bindings (typing.Optional[typing.Dict[str, typing.Any]], optional): A dictionary mapping variable names to values that should be available in the
            event handler's context. If None, no additional bindings are created.. Defaults to None.

    Example:
    .. code-block:: python
        a = ui.state(1)

        event = ui.vue_event(bindings={"a": a}, code=r'''()=> { a.value +=1}''')

        html.span(a)
        html.button("plus").on("click", event)
    """

    def __init__(
        self,
        *,
        code: str,
        bindings: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ):
        self.code = code
        self._bindings = bindings

    def copy_with_extends(self, extends: typing.Dict):
        raise NotImplementedError("VueEvent does not support extends")

    def event_type(self):
        return "vue"

    def _to_json_dict(self):
        data = super()._to_json_dict()
        data["type"] = self.event_type()
        if self._bindings:
            data["inputs"] = {
                name: {
                    "value": v._to_input_config()
                    if isinstance(v, CanInputMixin)
                    else v,
                    "type": v._to_event_input_type().value
                    if isinstance(v, CanInputMixin)
                    else InputBindingType.Data,
                }
                for name, v in self._bindings.items()
            }
        return data


vue_event = VueEvent



from __future__ import annotations
from typing import (
    Dict,
    Literal,
    Optional,
    Union,
    Sequence,
    Generic,
    TypeVar,
    cast,
    overload,
)
from enum import Enum

from instaui.components.component import Component
from instaui.vars.vfor_item import VForItem

from instaui.vars.mixin_types.element_binding import (
    ElementBindingMixin,
    ElementBindingProtocol,
)

_T = TypeVar("_T")


class VForArrayTypeEnum(Enum):
    CONST = "c"
    REF = "r"
    RANGE = "n"


class VFor(Component, Generic[_T]):
    def __init__(
        self,
        data: Union[Sequence[_T], ElementBindingProtocol, Dict[str, _T], _T],
        *,
        key: Union[Literal["item", "index"], str] = "index",
    ):
        """for loop component.

        Args:
            data (Union[Sequence[_T], ElementBindingMixin[List[_T]]]): data source.
            key (Union[Literal[&quot;item&quot;, &quot;index&quot;], str]]): key for each item. Defaults to 'index'.

        Examples:
        .. code-block:: python
            items = ui.state([1,2,3])

            with ui.vfor(items) as item:
                html.span(item)

            # object key
            items = ui.state([{"name": "Alice", "age": 20}, {"name": "Bob", "age": 30}])
            with ui.vfor(items, key=":item=>item.name") as item:
                html.span(item.name)

            # iter info
            items = ui.state({"a": 1, "b": 2, "c": 3})
            with ui.vfor(items) as item:
                info = ui.iter_info(item)

                html.span(info.dict_key)
                html.span(info.value)

            # range
            with ui.vfor.range(10) as i:
                html.paragraph(i)
        """

        super().__init__("vfor")
        self._data = data
        self._key = key
        self._num = None
        self._transition_group_setting = None

    def __enter__(self) -> _T:
        super().__enter__()
        return cast(_T, VForItem())

    def _set_num(self, num):
        self._num = num

    def transition_group(self, name="fade", tag: Optional[str] = None):
        self._transition_group_setting = {"name": name, "tag": tag}
        return self

    def _to_json_dict(self):
        data = super()._to_json_dict()

        if self._key is not None and self._key != "index":
            data["fkey"] = self._key

        if self._data is not None or self._num is not None:
            array_type = (
                VForArrayTypeEnum.REF
                if isinstance(self._data, ElementBindingMixin)
                else VForArrayTypeEnum.RANGE
                if self._num is not None
                else VForArrayTypeEnum.CONST
            )

            array_value = (
                self._data._to_element_binding_config()
                if isinstance(self._data, ElementBindingMixin)
                else self._num
                if self._num is not None
                else self._data
            )

            data["array"] = {"type": array_type, "value": array_value}

        if self._transition_group_setting is not None:
            data["tsGroup"] = {
                k: v for k, v in self._transition_group_setting.items() if v is not None
            }

        if self._slot_manager.has_slot():
            slot_data = self._slot_manager.get_slot(
                "default"
            )._to_items_container_config()
            data.update(slot_data)

        return data

    @overload
    @classmethod
    def range(cls, end: int) -> VFor[int]: ...

    @overload
    @classmethod
    def range(cls, end: ElementBindingProtocol) -> VFor[int]: ...

    @classmethod
    def range(cls, end: Union[int, ElementBindingProtocol]) -> VFor[int]:
        obj = cls(None)  # type: ignore

        num = (  # type: ignore
            end._to_element_binding_config()
            if isinstance(end, ElementBindingMixin)
            else end
        )

        obj._set_num(num)

        return obj  # type: ignore


def iter_info(item: _T) -> VForItem[_T]:
    return cast(VForItem[_T], item)



from __future__ import annotations
from typing import TYPE_CHECKING, Literal, Optional
from instaui.components.element import Element
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from . import _src_utils


if TYPE_CHECKING:
    from instaui.vars.types import TMaybeRef


class Video(Element):
    def __init__(
        self,
        src: Optional[TMaybeRef[str]] = None,
        *,
        controls: Optional[TMaybeRef[bool]] = None,
        autoplay: Optional[TMaybeRef[bool]] = None,
        loop: Optional[TMaybeRef[bool]] = None,
        muted: Optional[TMaybeRef[bool]] = None,
        playsinline: Optional[TMaybeRef[bool]] = None,
        poster: Optional[TMaybeRef[str]] = None,
        preload: Optional[TMaybeRef[Literal["auto", "metadata", "none"]]] = None,
    ):
        """
        Creates an HTML video element with configurable playback options.

        Args:
            src (Optional[TMaybeRef[str]]): The URL or path to the video file.
                - If a relative path is provided (e.g., "/xxx.mp4"), it will resolve
                to the `assets` directory in the application root.
                - Absolute URLs (e.g., "https://example.com/video.mp4") are also supported.
            controls (Optional[TMaybeRef[bool]]): Whether to display browser-native video controls.
            autoplay (Optional[TMaybeRef[bool]]): Whether the video should start playing automatically.
            loop (Optional[TMaybeRef[bool]]): Whether the video should loop after reaching the end.
            muted (Optional[TMaybeRef[bool]]): Whether the video should be muted by default.
            playsinline (Optional[TMaybeRef[bool]]): Whether the video should play inline on mobile devices.
            poster (Optional[TMaybeRef[str]]): URL to an image shown as a placeholder before playback starts.
            preload (Optional[TMaybeRef[Literal["auto", "metadata", "none"]]]): Hints how much video data should be preloaded.

        Example:
        .. code-block:: python
            from instaui import ui

            # Play a video from the assets directory
            ui.video("/xxx.mp4", controls=True, autoplay=False)

            # Play a remote video with a custom poster image
            ui.video("https://example.com/video.mp4", poster="/thumbnail.jpg", muted=True)
        """
        super().__init__("video")
        if src is not None:
            if isinstance(src, ElementBindingMixin):
                src = _src_utils.complete_src_computed(src)
            else:
                src = _src_utils.complete_src(src)
        self.props(
            {
                "src": src,
                "controls": controls,
                "autoplay": autoplay,
                "loop": loop,
                "muted": muted,
                "playsinline": playsinline,
                "poster": poster,
                "preload": preload,
            }
        )



from typing import Any, Union
from .ref import Ref
from .js_computed import JsComputed
from .vue_computed import VueComputed
from .web_computed import WebComputed


from ._types import _T_Value
from .mixin_types.element_binding import ElementBindingMixin


TRefOrComputed = Union[
    Ref[_T_Value],
    VueComputed,
    JsComputed,
    WebComputed[Any, _T_Value],
]
TMaybeRef = Union[
    ElementBindingMixin[_T_Value],
    WebComputed,
    VueComputed,
    JsComputed,
    _T_Value,
]



from typing import Union
from pathlib import Path
import inspect
from instaui.runtime.context import get_context


def create_server(
    debug: bool = False,
    use_gzip: Union[int, bool] = True,
):
    """
    Create a new server instance.

    Note: When deploy to production, you need to set `debug=False`

    Args:
        debug (bool): Whether to run in debug mode. In debug mode, it has the development hot-reload feature. Defaults to False.
        use_gzip (Union[int, bool], optional):  Whether to use gzip compression. If an integer is provided, it will be used as the minimum response size for compression. If True, the default minimum size of 500 bytes will be used. If False, compression will not be used. Defaults to True.
    """
    from instaui.fastapi_server.server import Server

    context = get_context()
    context._app_mode = "web"
    context._debug_mode = debug
    caller_file_path = get_caller_file_path()

    return Server(use_gzip=use_gzip, caller_file_path=caller_file_path)


def get_caller_file_path() -> Path:
    frame = inspect.currentframe().f_back.f_back  # type: ignore
    assert frame is not None
    file_path = inspect.getfile(frame)
    return Path(file_path).resolve()



from __future__ import annotations
from collections import deque
from datetime import datetime
import importlib.util
from pathlib import Path
import typing
import jinja2
import inspect
from dataclasses import dataclass, field


def build_routes_from_files(
    folder_path: typing.Union[str, Path] = "pages",
    module_name: str = "_routes",
    route_config_var_name: str = "_route_config",
):
    global_args = _GlobalArgs(
        base_folder_path=_utils.get_caller_path().parent / Path(folder_path),
        module_name=module_name,
        route_config_var_name=route_config_var_name,
    )

    root = _model_utils.create_root(global_args)
    _code_gen.generate_router_file(root)

    print(f"Build _routes from files in {global_args.base_folder_path}...")


@dataclass
class _GlobalArgs:
    base_folder_path: Path
    module_name: str
    route_config_var_name: str


class _model_utils:
    @dataclass
    class FileRouteInfo:
        file: Path
        base_folder: Path
        global_args: _GlobalArgs
        children: typing.List[_model_utils.FileRouteInfo] = field(default_factory=list)
        path: str = field(init=False, default="")
        name: str = field(init=False, default="")
        fn_path: typing.Optional[str] = field(init=False, default=None)

        params: str = field(init=False, default="")
        meta: typing.Dict = field(init=False, default_factory=dict)

        def __post_init__(self):
            self.params = self._extract_params()
            self.meta = self._extract_meta()
            self.path, self.name = self._extra_path_name()

            if self.file.is_file():
                self.fn_path = ".".join(
                    self.file.relative_to(self.base_folder).with_suffix("").parts
                )

        def is_index_file(self):
            return self.file.is_file() and self.file.stem == "index"

        def change_sync_index_info(self, index_info: _model_utils.FileRouteInfo):
            self.fn_path = index_info.fn_path
            self.path = self.path + index_info.params
            self.meta = index_info.meta

        def _extract_params(self):
            if self.file.is_file():
                route_config = _module_utils.get_module_getter(self.file)(
                    self.global_args.route_config_var_name
                )
                if route_config:
                    if "params" in route_config:
                        return route_config["params"]

            return ""

        def _extract_meta(self):
            if self.file.is_file():
                route_config = _module_utils.get_module_getter(self.file)(
                    self.global_args.route_config_var_name
                )

                if route_config:
                    if "meta" in route_config:
                        return route_config["meta"]

            return {}

        def _extra_path_name(self):
            name_parts = list(
                self.file.relative_to(self.base_folder).with_suffix("").parts
            )

            is_root = len(name_parts) == 1

            path = self.file.stem
            if path == "index":
                path = ""

            name = ".".join(name_parts)

            if is_root:
                path = "/" + path

            if self.params:
                path += self.params

            return path, name

        def import_code(self):
            if not self.fn_path:
                return ""

            return f"from .{self.fn_path.replace(' ', '_')} import main as {self.main_fn_name()}"

        def main_fn_name(self):
            if not self.fn_path:
                return ""
            return self.name.replace(".", "_").replace(" ", "_")

    @dataclass
    class FileRouteRoot:
        folder: str
        module_name: str
        infos: list[_model_utils.FileRouteInfo] = field(default_factory=list)

    @staticmethod
    def create_root(global_args: _GlobalArgs) -> FileRouteRoot:
        base_folder = Path(global_args.base_folder_path)
        infos = _model_utils._create_route_info(base_folder, global_args)
        return _model_utils.FileRouteRoot(
            folder=str(base_folder), module_name=global_args.module_name, infos=infos
        )

    @staticmethod
    def _create_route_info(
        base_folder: Path, global_args: _GlobalArgs
    ) -> typing.List[FileRouteInfo]:
        result: typing.List[_model_utils.FileRouteInfo] = []

        stack: deque[
            typing.Tuple[typing.Optional[_model_utils.FileRouteInfo], Path]
        ] = deque()
        stack.extendleft((None, path) for path in base_folder.iterdir())

        while stack:
            parent_info, item = stack.pop()
            is_dir = item.is_dir()

            if item.stem.startswith("_"):
                continue

            if is_dir:
                folder_info = _model_utils.FileRouteInfo(
                    file=item, base_folder=base_folder, global_args=global_args
                )
                infos = ((folder_info, path) for path in item.iterdir())
                stack.extendleft(infos)

                if parent_info is None:
                    result.append(folder_info)
                else:
                    parent_info.children.append(folder_info)
                continue

            if item.suffix != ".py":
                continue

            file_info = _model_utils.FileRouteInfo(
                file=item, base_folder=base_folder, global_args=global_args
            )

            if parent_info is None:
                result.append(file_info)
            else:
                if file_info.is_index_file():
                    parent_info.change_sync_index_info(file_info)

                else:
                    parent_info.children.append(file_info)

        return result

    @staticmethod
    def iter_route_info(infos: typing.List[FileRouteInfo]):
        stack: typing.List[_model_utils.FileRouteInfo] = []
        stack.extend(infos)

        while stack:
            info = stack.pop()
            stack.extend(info.children)
            yield info


class _code_gen:
    _env = jinja2.Environment(
        loader=jinja2.PackageLoader("instaui.spa_router", "templates"),
    )

    @dataclass
    class TemplateModel:
        update_time: datetime = field(default_factory=datetime.now)
        route_names: typing.List[str] = field(default_factory=list)
        routes: typing.List[_model_utils.FileRouteInfo] = field(default_factory=list)

        def get_all_main_import(self):
            return [
                info.import_code() for info in _model_utils.iter_route_info(self.routes)
            ]

    @staticmethod
    def generate_router_file(root: _model_utils.FileRouteRoot):
        _template = _code_gen._env.get_template("page_routes")

        template_model = _code_gen.TemplateModel(
            route_names=_code_gen._extract_all_route_names(root), routes=root.infos
        )

        code = _template.render(model=template_model)
        Path(root.folder).joinpath(f"{root.module_name}.py").write_text(
            code, encoding="utf-8"
        )

    @staticmethod
    def _extract_all_route_names(root: _model_utils.FileRouteRoot):
        return [
            info.name
            for info in _model_utils.iter_route_info(root.infos)
            if info.fn_path
        ]


class _module_utils:
    @staticmethod
    def get_module_getter(path: Path):
        if not isinstance(path, Path):
            raise ValueError("Expected a Path object")

        if not path.exists():
            raise FileNotFoundError(f"The file {path} does not exist.")

        module_name = path.stem
        module_path = str(path.absolute())

        spec = importlib.util.spec_from_file_location(module_name, module_path)
        if spec is None:
            raise ImportError(f"Cannot create a module spec for {module_path}")

        module = importlib.util.module_from_spec(spec)

        try:
            spec.loader.exec_module(module)  # type: ignore
        except Exception as e:
            raise ImportError(f"Failed to import {module_path}: {e}")

        def getter_fn(var_name: str):
            return getattr(module, var_name, None)

        return getter_fn


class _utils:
    @staticmethod
    def get_caller_path():
        current_frame = inspect.currentframe()
        try:
            caller_frame = current_frame.f_back.f_back  # type: ignore
            filename = caller_frame.f_code.co_filename  # type: ignore
            return Path(filename)
        finally:
            del current_frame



from __future__ import annotations
import inspect
from typing import (
    Any,
    Callable,
    Dict,
    Generic,
    Optional,
    Sequence,
    TypeVar,
)
from typing_extensions import ParamSpec

from instaui.common.jsonable import Jsonable
from instaui.runtime._app import get_app_slot, get_current_scope
from instaui.handlers import watch_handler

from instaui.vars._types import InputBindingType
from instaui.vars.path_var import PathVar
from instaui.vars.mixin_types.py_binding import (
    CanInputMixin,
    CanOutputMixin,
    outputs_to_config,
)
from instaui.vars.mixin_types.element_binding import ElementBindingMixin
from instaui.vars.mixin_types.pathable import CanPathPropMixin
from instaui.vars.mixin_types.str_format_binding import StrFormatBindingMixin
from instaui.vars.mixin_types.observable import ObservableMixin
from instaui.vars.mixin_types.common_type import TObservableInput
from instaui._helper import observable_helper
from instaui import pre_setup as _pre_setup

_SYNC_TYPE = "sync"
_ASYNC_TYPE = "async"

P = ParamSpec("P")
R = TypeVar("R")


class WebComputed(
    Jsonable,
    PathVar,
    CanInputMixin,
    ObservableMixin,
    CanPathPropMixin,
    StrFormatBindingMixin,
    ElementBindingMixin[R],
    Generic[P, R],
):
    def __init__(
        self,
        func: Callable[P, R],
        inputs: Optional[Sequence[TObservableInput]] = None,
        extend_outputs: Optional[Sequence[CanOutputMixin]] = None,
        init_value: Optional[R] = None,
        evaluating: Optional[CanOutputMixin] = None,
        deep_compare_on_input: bool = False,
        pre_setup: Optional[Dict] = None,
        debug_info: Optional[Dict] = None,
    ) -> None:
        scope = get_current_scope()

        self.__register_info = scope.register_computed_task(self)
        self._org_inputs = inputs or []

        self._inputs, self._is_slient_inputs, self._is_data = (
            observable_helper.analyze_observable_inputs(list(inputs or []))
        )
        self._outputs = outputs_to_config([output for output in extend_outputs or []])
        self._fn = func
        self._init_value = init_value
        self._deep_compare_on_input = deep_compare_on_input
        self._pre_setup = pre_setup
        if evaluating is not None:
            self._pre_setup = self._pre_setup or {}
            self._pre_setup = {**self._pre_setup, evaluating: True}

        if debug_info is not None:
            self.debug = debug_info

    def __call__(self, *args: P.args, **kwargs: P.kwargs) -> R:
        return self._fn(*args, **kwargs)

    def _to_json_dict(self):
        data = super()._to_json_dict()

        app = get_app_slot()

        hkey = watch_handler.create_handler_key(
            page_path=app.page_path,
            handler=self._fn,
        )

        watch_handler.register_handler(hkey, self._fn, len(self._outputs) + 1)

        data["id"] = self.__register_info.var_id

        if self._inputs:
            data["inputs"] = self._inputs

        if self._outputs:
            data["outputs"] = self._outputs

        if sum(self._is_slient_inputs) > 0:
            data["slient"] = self._is_slient_inputs

        if sum(self._is_data) > 0:
            data["data"] = self._is_data

        data["fType"] = (
            _ASYNC_TYPE if inspect.iscoroutinefunction(self._fn) else _SYNC_TYPE
        )
        data["key"] = hkey
        if self._init_value is not None:
            data["init"] = self._init_value

        if self._deep_compare_on_input is not False:
            data["deepEqOnInput"] = 1
        if self._pre_setup:
            data["preSetup"] = _pre_setup.convert_config(self._pre_setup)

        return data

    def __to_binding_config(self):
        return {
            "id": self.__register_info.var_id,
            "sid": self.__register_info.scope_id,
        }

    def _to_pathable_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_path_prop_binding_config(self) -> Dict:
        return self.__to_binding_config()

    def _to_input_config(self):
        return self.__to_binding_config()

    def _to_element_binding_config(self):
        return self.__to_binding_config()

    def _to_observable_config(self):
        return self.__to_binding_config()

    def _to_event_input_type(self) -> InputBindingType:
        return InputBindingType.Ref


def web_computed(
    *,
    inputs: Optional[Sequence] = None,
    extend_outputs: Optional[Sequence] = None,
    init_value: Optional[Any] = None,
    evaluating: Optional[Any] = None,
    deep_compare_on_input: bool = False,
    pre_setup: Optional[Dict] = None,
    debug_info: Optional[Dict] = None,
):
    """
    Creates a computed property decorator for reactive programming with dependency tracking.

    This decorator factory wraps functions to create reactive computed properties that:
    - Automatically re-evaluate when dependencies (inputs) change
    - Cache results for performance optimization
    - Support both synchronous and asynchronous computation patterns

    Args:
        inputs (Optional[Sequence], optional): Collection of reactive sources that trigger recomputation
                                   when changed. These can be state objects or other computed properties.
        extend_outputs (Optional[Sequence], optional):  Additional outputs to notify when this computed value updates.
        init_value (Optional[Any], optional): Initial value to return before first successful evaluation.
        evaluating (Optional[Any], optional): Temporary value returned during asynchronous computation.
        pre_setup (typing.Optional[typing.Dict], optional): A dictionary of pre-setup actions to be executed before the event executes.

    # Example:
    .. code-block:: python
        from instaui import ui,html

        a = ui.state(0)

        @ui.computed(inputs=[a])
        def plus_one(a):
            return a + 1

        html.number(a)
        ui.label(plus_one)
    """

    def wrapper(func: Callable[P, R]):
        return WebComputed(
            func,
            inputs=inputs,
            extend_outputs=extend_outputs,
            init_value=init_value,
            evaluating=evaluating,
            deep_compare_on_input=deep_compare_on_input,
            pre_setup=pre_setup,
            debug_info=debug_info,
        )

    return wrapper


TComputed = WebComputed



from typing import Callable
from instaui.launch_collector import get_launch_collector, PageInfo


def page(path: str = "/"):
    """Register a page route.

    Args:
        path (str): The route path.
    """

    def wrapper(func: Callable):
        get_launch_collector().register_page(PageInfo(path, func))
        return func

    return wrapper



from __future__ import annotations
import typing
from instaui import html
from instaui.components.element import Element
from instaui.vars.types import TMaybeRef


class RouterLink(Element):
    def __init__(self, text: TMaybeRef[str], *, to: str):
        super().__init__("router-link")

        self.props({"to": to})

        if text is not None:
            with self.add_slot("default"):
                html.span(text)

    @classmethod
    def by_name(
        cls,
        text: TMaybeRef[str],
        *,
        name: str,
        params: typing.Optional[typing.Dict[str, typing.Any]] = None,
    ) -> RouterLink:
        to: typing.Dict = {"name": name}
        if params:
            to["params"] = params

        return cls(text, to=to)  # type: ignore


class RouterView(Element):
    def __init__(self):
        super().__init__("router-view")



from __future__ import annotations
from typing import Any, Dict
from instaui.components.component import Component
from instaui.vars.types import TMaybeRef
from instaui.vars.mixin_types.element_binding import ElementBindingMixin


class Content(Component):
    def __init__(self, content: TMaybeRef[Any]):
        """Content to be displayed on the page, typically used for pure text content within slots.

        Args:
            content (TMaybeRef[Any]): The textual content to display.

        Examples:
        .. code-block:: python
            with html.div():
                ui.content("Hello, world!")
        """
        super().__init__("content")
        self._content = content

    def _to_json_dict(self) -> Dict:
        data = super()._to_json_dict()
        props = {}
        data["props"] = props

        if isinstance(self._content, ElementBindingMixin):
            props["content"] = self._content._to_element_binding_config()
            props["r"] = 1
        else:
            props["content"] = self._content

        return data



from typing import Any, Iterable, Sequence, Union
from instaui.common.jsonable import Jsonable


class PatchSetRecord:
    def __init__(self, *, path: Sequence[Union[str, int]], value: Any):
        self.path = list(path)
        self.value = value


class PatchSet(Jsonable):
    def __init__(self, records: Iterable[PatchSetRecord]):
        self._records = list(records)

    def patch_set(self, *, path: Sequence[Union[str, int]], value: Any):
        return PatchSet([*self._records, PatchSetRecord(path=path, value=value)])

    def _to_json_dict(self):
        return [[r.path, r.value] for r in self._records]


def patch_set(*, path: Sequence[Union[str, int]], value: Any):
    """
    Generates a patch to specify how data should be updated.

    This function is typically used within the return value of functions
    used in `ui.event` or `ui.watch`, indicating the way data modifications are applied.

    Args:
        path (Sequence[Union[str, int]]): A sequence representing the path to the item that needs to be updated.
                                           Each element can either be a string (for dictionary keys) or an integer
                                           (for list indices).
        value (Any): The new value to set at the specified path. Can be of any type.

    Example:
    .. code-block:: python

        data = ui.state(
            {
                "v1": ["a", "b", "c"],
                "v2": ["x", "y", "z"],
            }
        )

        @ui.event(outputs=[data])
        def update_data():
            # update the second element of "v1" to "foo"
            return ui.patch_set(path=["v1", 1], value="foo")

        html.button("update data").on_click(update_data)
        ui.label(data)
    """

    return PatchSet([PatchSetRecord(path=path, value=value)])
